// Copyright 2013-2014, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.

package com.sas.deployment
import groovy.transform.InheritConstructors

import com.sas.gate.CodeAccessor
import com.sas.gate.api.PipeCode
import com.sas.gate.api.template.*
import org.codehaus.groovy.runtime.StackTraceUtils

/**
   <overview>Deployment code for SAS Compliance Solutions Mid-Tier (antimnylndin)</overview>
   <p>
      This code configures the deployment of 12byte using what corresponds to
      the following Ant targets:
      <ul class='nodot'>
         <li>configure</li>
         <li>buildWebapps</li>
         <li>deployWebapps</li>
         <li>loadContent</li>
         <li>unconfigure</li>
         <li>updatePasswords</li>
         <li>updateConfigure</li>
      </ul>
   </p>

   @since      14w14
   @version    15w04
   @author     Bruce Bryan
   @see        #SOURCE#
*/

@InheritConstructors
class ANTIMNYLNDIN extends CodeAccessor
{
   // ===================================================== GLOBAL

   /**
      Identifier

      @param id <type>String</type>: The 12byte name
   */
   String id = 'ANTIMNYLNDIN'

   /**
      Callback for hook: <hook>properties &gt; init</hook>

      @param t__properties_get   <type>ArrayList&lt;String&gt;</type>: get object properties on each dependson property

   <ul class='nodot'>
      <li>webappsrv</li>
      <li>webinfpltfm</li>
      <li>casemgmtmid</li>
      <li>antimnycmnc</li>
   </ul>
   */
   def properties_init =
   {
      pass.with
      {

         // Simple Strings
         _releaseNumber                   = api.'antimnylndin.release.number'
         _offeringLegalName               = api.'antimnylndin.offering.legal.name'
         _normalizedOfferingLegalName     = api.'antimnylndin.normalized.offering.name'
         _normalizedShortLegalName        = api.'antimnylndin.normalized.short.legal.name'
         _applicationName                 = api.'antimnylndin.normalized.component.name'
         _rootAppName                     = 'sas.financialservices.aml'
         _shortLegalName                  = api.'antimnylndin.short.legal.name'
         _softwareComponentName           = api.'antimnylndin.software.component.name'
         // Config Paths
         _sasConfigDir                    = api.'config.lev.web.applications.dir'
         _configDir                       = canonical("$pass._sasConfigDir/$pass._normalizedOfferingLegalName")
         _configInstallDir                = canonical("$_configDir/install")
         _configMetadataDir               = canonical("$_configInstallDir/metadata")

         // Install Paths (Dirs and Files)
         _installDir                      = api.'antimnylndin.install.dir'
         _installConfigDir                = api.'antimnylndin.install.config.dir'
         _installMetadataDir              = canonical("$pass._installConfigDir/Deployment/Metadata")
         _installPackagesDir              = canonical("$pass._installConfigDir/Deployment/Packages")

         // Other
         _contextName                     = api.'antimnylndin.context.name'
         _davPath                         = "Products/$pass._normalizedOfferingLegalName"
         _davFolder                       = "/SAS$pass._normalizedShortLegalName"
         _webappsrvHost                   = api.'antimnylndin.webappsrv.host'
         _webappsrvPort                   = api.'antimnylndin.webappsrv.port'
         _webappsrvProtocol               = api.'antimnylndin.webappsrv.protocol'
         _webappsrvUrl                    = api.'vfabrcersvrc.server.url'
         _fcfUrl                          = "$pass._webappsrvUrl$pass._contextName"
         _segkcDatasource                 = api.'antimnylndin.dbms.segkc.datasource.name'
         _dbmsEngine                      = api.'antimnylndin.dbms.engine'
         _classPath                       = [
                                                 db2       : api.'antimnylndin.dbms.segkc.db2_driver' + ';' + api.'antimnylndin.dbms.segkc.db2_license',
                                                 oracle    : api.'antimnylndin.dbms.segkc.oracle_driver',
                                                 sqlsvr    : api.'antimnylndin.dbms.segkc.sqlserver_driver',
                                                 odbc      : api.'antimnylndin.dbms.segkc.sqlserver_driver',
                                                 postgres  : api.'antimnylndin.dbms.segkc.postgresql_driver',
                                                 teradata  : api.'antimnylndin.dbms.segkc.teradata_driver'
                                             ]."$pass._dbmsEngine"
         _validationQuery                 = [
                                                 db2       : 'select 1 from sysibm.sysdummy1',
                                                 oracle    : 'select 1 from dual',
                                                 sqlsvr    : 'select 1 from sysobjects',
                                                 odbc      : 'select 1 from sysobjects',
                                                 postgres  : 'select 1',
                                                 teradata  : 'select 1'
                                             ]."$pass._dbmsEngine"
         _jndiName                        = "sas/jdbc/$pass._segkcDatasource"


         // Template APIs
         t__properties_get =
         [
            'antimnylndin.dependson.webappsrv.fqid',
            'antimnylndin.dependson.webinfpltfm.fqid',
            'antimnylndin.dependson.casemgmtmid.fqid',
            'antimnylndin.dependson.antimnycmnc.fqid'
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param filters <type>Map&lt;String&gt;</type>: tokens for string substitution
      @param t__setobjectproperties    <type>ArrayList&lt;String&gt;</type>: Properties that will be stored in the configured objects text store
   */
   def properties_post =
   {
      // Simple Strings
      String appServerName                = api.'oma.appserver.servercontext.name'
      String applicationFolder            = api.'antimnylndin.application.folder'
      String cfgLevWebLogsDir             = api.'config.lev.web.logs.dir'
      String classIdentifier              = api.'antimnylndin.configobj.classid'
      String configobjName                = api.'antimnylndin.configobj.name'
      String contextRoot                  = pass._contextName
      String productDescription           = "SAS $pass._softwareComponentName"
      String displayName                  = "SAS $pass._shortLegalName $pass._releaseNumber"
      String majorNumber                  = api.'antimnylndin.major.number'
      String minorNumber                  = api.'antimnylndin.minor.number'
      String monthLen                     = api.'antimnycmnc.agp.month.length.num'
      String repositoryName               = api.'oma.repository.foundation.name'
      String legalName                    = api.'antimnylndin.legal.name'
      String schemaName                   = api.'antimnylndin.dbms.schema'
      String segmentId                    = api.'antimnylndin.ecm.source_system_cd'
      String trustedUserName              = api.'oma.person.trustusr.person.name'
      String webHost                      = api.'antimnylndin.webappsrv.host'
      String webPort                      = api.'antimnylndin.webappsrv.port'
      String webProtocol                  = api.'antimnylndin.webappsrv.protocol'
      String stpMetadataFolder            = "${api.'antimnylndin.antimnycmnc.application.folder'}/Application SAS Code"
      String ecmSoftwareComponentName     = api.'antimnylndin.casemgmtmid.software.component.name'
      String stpSoftwareComponentName     = 'Stored Process Web App 9.4'
      String dbmsEngine                   = [
                                                 db2       : 'db2',
                                                 oracle    : 'oracle',
                                                 sqlsvr    : 'sqlserver',
                                                 odbc      : 'sqlserver',
                                                 postgres  : 'postgres',
                                                 teradata  : 'teradata'
                                             ]."$pass._dbmsEngine"

      Map serverLogDirs                   = api.serverLogDirs()
      String databaseSid                  = (api.'antimnylndin.dbms.engine' == 'oracle') ? "${api.'antimnycmnc.dbms.oracle.sid'}" : ''

      // Properties
      api.createProperty('antimnylndin.log.dir.html', serverLogDirs.html)
      if (databaseSid) api.createProperty('antimnylndin.dbms.oracle.sid', databaseSid)

      if (t__debug){
         assert classIdentifier == 'antimnylndin' : "Class identifier does not match antimnylndin"
         assert majorNumber == '6' : "Release major number does not equal 6"
         assert minorNumber == '3' : "Release minor number does not equal 3"
         assert legalName == 'SAS Compliance Solutions Mid-Tier' : "Product legal name does not match: SAS Compliance Solutions Mid-Tier"
      }

      pass.with
      {
         _serverName                      = api.serverName()
         _serverSoftwareComponentName     = api.'antimnycmnc.software.component.name'
         _filters =
         [
            'aml.dav.path'                : _davPath,
            'cdd.dav.path'                : _davPath,
            'dav.path'                    : _davPath,
            'aml.software.component.name' : _softwareComponentName,
            'business.unit.name'          : api.'antimnylndin.business.unit.name',
            'class.identifier'            : classIdentifier,
            'component.name'              : _softwareComponentName,
            'dbms.engine'                 : dbmsEngine,
            'ecm.software.component.name' : ecmSoftwareComponentName,
            'fcf.software.component.name' : _softwareComponentName,
            'major.number'                : majorNumber,
            'minor.number'                : minorNumber,
            'month.length'                : monthLen,
            'normalized.offering.name'    : _normalizedOfferingLegalName,
            'product.description'         : productDescription,
            'product.name'                : legalName,
            'release.number'              : _releaseNumber,
            'segkc.datasource'            : _segkcDatasource,
            'system.product.folder'       : applicationFolder,
            'war.context.name'            : _contextName,
            'webappsrv.host'              : webHost,
            'webappsrv.port'              : webPort,
            'webappsrv.protocol'          : webProtocol,
            'segment_id'                  : segmentId,
            'stp.software.component.name' : stpSoftwareComponentName,
            'stp.metadata.folder'         : stpMetadataFolder,
            'war.archive.name'            : _rootAppName,
         ]

         t__setobjectproperties =
         [
            properties:
            [
               'antimnylndin.context.name',
               'antimnylndin.dbms.engine',
               'antimnylndin.dbms.host',
               'antimnylndin.dbms.port',
               'antimnylndin.dbms.segkc.schema',
               'antimnylndin.dbms.segkc.userid',
               'antimnylndin.dbms.type',
               'antimnylndin.ecm.source_system_cd',
               'antimnylndin.release.number'
            ]
         ]
         switch(_dbmsEngine)
         {
         case 'db2' :
            t__setobjectproperties.properties +=
               [
                  'antimnylndin.dbms.database.name',
                  'antimnylndin.dbms.segkc.db2_license'
               ]
            break
         case 'oracle' :
            t__setobjectproperties.properties +=
            [
               'antimnylndin.dbms.oracle.service.name',
               'antimnylndin.dbms.oracle.sid'
            ]
            break
         default :
            t__setobjectproperties.properties +=
            [
                  'antimnylndin.dbms.database.name'
            ]
            break
         }
      }

      api.createProperty('antimnylndin.webappsrv.name', "$pass._serverName")

   }

   def properties_assert =
   {
      pass.with {
         assert _releaseNumber == "6.3" : "Release number does not match 6.3"
         assert _offeringLegalName == 'SAS Compliance Solutions' : "Offering legal name does not match: SAS Compliance Solutions"
         assert _shortLegalName == 'Compliance Solutions Mid' : "Short Legal name does not match: Compliance Solutions Mid"
         assert _softwareComponentName == 'Compliance Solutions Mid 6.3' : "Software component name does not match: Compliance Solutions Mid 6.3"
         assert _configDir =~ "SASComplianceSolutions" : "Config dir does not contain: SASComplianceSolutions"
         assert _configInstallDir =~ "install" : "Config dir does not contain install"
         assert _configMetadataDir =~ "metadata" : "Config dir does not contain metadata"
      }
   }

   // ===================================================== TARGET: configure

   /**
      Target: configure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def configure()
   {
      //api.stop()

      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      ConfigureTemplate t = new ConfigureTemplate(this)

      t['migration > init']         .add(configure__migration_init)
      t['properties > init']        .add(properties_init)
      t['properties > post']        .add(properties_post)
      t['properties > post']        .add(configure__migrate_properties_post).when('t__migrate')
      t['properties > post']        .add(properties_assert).when('t__debug')
      t['init']                     .add(configure__init)
      t['loadmetadata > migration > init']   .add(configure__loadmetadata_migration_init).when('t__migrate')
      t['docparts > init']          .add(configure__docparts_init).when('!t__migrate')
      t['docparts > migration']     .add(configure__docparts_migration).when('t__migrate')

      t.execute()

      //api.stop()
   }

   def configure__migration_init =
   {
      String shortLegalName = api.'antimnylndin.short.legal.name'
      String maintenanceFrom = getProperty('antimnylndin.maintenance.from.version', '')
      String oldApplicationFolder = "$shortLegalName $maintenanceFrom"
      String releaseNumber = api.'antimnylndin.release.number'

      pass.with {
         _updateMetadata =
         t__migrate = getProperty('antimnylndin.migration.is_enabled', '0') { toBoolean(it) }
         if (t__migrate)
         {
            _migrationFrom = getProperty('antimnylndin.migration.from.group', '_51_and_earlier')

            if (_migrationFrom == '_62_and_later')
            {
               if ( maintenanceFrom != releaseNumber )
               {
                  t__searchandcall =
                  [
                     fqids:
                     [
                        [
                           actions: [ __template.t__deletemetadata ],
                           type:    'Tree',
                           search:  "@Name = '$oldApplicationFolder'"
                        ]
                     ]
                  ]
               }
            }
            else
            {
               t__removemetadatabypath =
               [
                  paths:
                  [
                     [
                        deleteContents : true,
                        objectPath : '/System/Applications/SAS Financial Crimes Foundation'
                     ]
                  ]
               ]
            }

            switch (api.'config.migration.sas.version')
            {
            case '9.2':
               _oldContextName = ''
               break

            case '9.3':
               _oldContextName = '/FinancialCrimesFdnMid'
               break

            case '9.4':
               String fromVersion = api.'antimnylndin.maintenance.from.version'
               _oldContextName = ((fromVersion == '6.1')
               ? '/SASFinancialCrimesFdnMid'
               : _contextName)
               break

            default:
               break
            }
         }
      }
   }

   def configure__migrate_properties_post =
   {
      pass.with {
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]
      }
   }

   /**
      @param t__createdirectories <type>Map&lt;String&gt;</type>: directories to be created in configure target<br>
            <indent><key>prefix</key>- <type>String</type>: base directory to create<br>
            <key>entries</key>- <type>ArrayList&lt;String&gt;</type>: directories and subdirectories to be created<br></indent>
      <pre>
   SAS_CONFIG/Lev1/Web/Applications/SASComplianceSolutions
   `-- install
       |-- Contents
       |-- content
       |-- customProperties
       |-- menu
       |-- metadata
       |-- uidef
       `-- workflow</pre>
      @param t__copyfiles         <type>ArrayList&lt;Map&gt;</type>: files to copy and conditions for the copy<br>
            <indent><key>src</key>- <type>String</type>: Source directory to copy from<br>
            <key>include</key>- <type>String</type>: Regex pattern of file(s) to copy<br>
            <key>dst</key>- <type>String</type>: Destination directory for files to be copied to<br>
            <key>strip</key>- <type>String</type> <kind>optional</kind>: Remove extension form file name i.e., remove .orig extension<br>
            <key>tokens</key>- <type>String</type> <kind>optional</kind>: filterset to use for token replacement<br></indent>
      @param t__loadmetadata_load   <type>String</type>: Fully qualified path to the token substituted appxml file
      @param t__datasource_create   <type>ArrayList&lt;Map&gt;</type>: Creates a JDBC datasource<br>
   */
   def configure__init =
   {

      pass.with {
         t__createdirectories =
         [
            prefix: _configInstallDir,
            entries:
            [
               'content',
               'Contents',
               'customProperties',
               'menu',
               'metadata',
               'uidef',
               'workflow',
            ]
         ]

         t__copyfiles =
         [
            [  src:        _installMetadataDir,
               include:    ~/.*\.orig/,
               dst:        _configMetadataDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.locprop/,
               exclude:    ~/.*\.orig/,
               dst:        _configMetadataDir,
            ],
         ]

         t__loadmetadata_load = canonical("$pass._configMetadataDir/antimnylndin.appxml")
// TODO: research txType and see why it is necessary May be the case of not being able to
//       deleteDatasource in unconfigure.
         t__datasource_create =
         [
            adminUser         : api.'antimnylndin.dbms.segkc.userid',
            adminPW           : api.decodeSASstring(api.'antimnylndin.dbms.segkc.passwd'),
            classPath         : _classPath,
            connectionURL     : api.'antimnylndin.dbms.segkc.url',
            datasourceName    : _segkcDatasource,
            driver            : api.'antimnylndin.dbms.segkc.driver_class',
            jndiName          : _jndiName,
            servername        : _serverName,
            validationQuery   : _validationQuery,
            txType:           'None'
         ]

         t__importpackages_run =
         [
            packages:
            [
               [
               name:          'ComplianceSolutions',
               packagesDir:   _installPackagesDir,
               folder:        '/System/Applications/SAS Social Network Analysis/Social Network Analysis 6.2'
               ],
            ]
         ]
      }
   }

   def configure__loadmetadata_migration_init =
   {
      String applicationSubFolder = api.'antimnylndin.application.folder'

      pass.with {
         t__loadmetadata_migration_update =
         [
            file: canonical("$_configMetadataDir/antimnylndin.appxml"),
            folder: applicationSubFolder,
            applicationname: _softwareComponentName
         ]

         if (_oldContextName)
         {
            String search = "@Service ? '" + _oldContextName + "'" + ']'
            String host = api.'antimnylndin.webappsrv.host'
            String port = api.'antimnylndin.webappsrv.port'
            String protocol = api.'antimnylndin.webappsrv.protocol'

            if (_migrationFrom == '_62_and_later')
            {
               Map params =
               [
                  actions: [ __template.t__updatetcpipconnection ],
                  type:    'TCPIPConnection',
                  search:  search,
                  args:
                  [
                     message: "Updating TCP/IP connection: $protocol://$host:$port/$search",
                     host:    host,
                     port:    port,
                     protocol: protocol
                  ]
               ]
               t__searchandcall.fqids += params
            }
            else
            {
               t__searchandcall =
               [
                  actions: [ __template.t__updatetcpipconnection ],
                  type:    'TCPIPConnection',
                  search:  search,
                  args:
                  [
                     message: "Updating TCP/IP connection: $protocol://$host:$port/$search",
                     host:    host,
                     port:    port,
                     protocol: protocol
                  ]
               ]
            }
         }
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def configure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.antimnylndin.head',
                  value:api.'instructions_antimnylndin_head.txt']
               ],
               [replace:[
                  token:'instructions.antimnylndin.details.table.head',
                  value:api.'instructions_antimnylndin_details_table_head.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_doc.txt']
               ],
               [replace:[
                  token:'instructions.antimnylndin.details.table.tail',
                  value:api.'instructions_antimnylndin_details_table_tail.txt']
               ]
            ]
         ]
      }
   }

   def configure__docparts_migration =
   {
      pass.with {
         String text =(_migrationFrom == '_51_and_earlier'
         ? api.'instructions_antimnylndin_migration_51_and_earlier.txt'
         : (api.'antimnylndin.maintenance.from.version' == '5.2'
           ? api.'instructions_antimnylndin_migration_52.txt'
           : api.'instructions_antimnylndin_migration.txt'))

         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.antimnylndin.head',
                  value: text]
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_doc.txt']
               ],
               [replace:[
                  token:'instructions.antimnylndin.details.table.tail',
                  value:api.'instructions_antimnylndin_details_table_tail.txt']
               ]
            ]
         ]
      }
   }

   def sasjobs_createuser =
   {
      api.createUserAddToGroups(pass.t__sasjobs_run)
   }

   def sasjobs_removeuser =
   {
      api.removeUserFromGroupsAndDelete(pass.t__sasjobs_run)
   }

   // ===================================================== TARGET: buildWebapps

   /**
      Target: buildWebapps

      @see com.sas.gate.api.template.BuildWebappsTemplate
   */
   def buildWebapps()
   {
      //api.stop()

      String maintenanceFrom  = getProperty('antimnylndin.maintenance.from.version', '')

      if (maintenanceFrom)
      {
         String fromVersion         = api.baseVersion(maintenanceFrom)
         String maintenanceTo = getProperty('antimnylndin.maintenance.to.version', '')
         String toVersion           = api.baseVersion(maintenanceTo)
         if (fromVersion != toVersion)
         {
            BuildWebappsTemplate t = new BuildWebappsTemplate(this)

            t.define(
            [
               'properties > pre init',
            ])

            t['properties > pre']         .add(updateconfigure__pre_init)
            t['properties > init']        .add(properties_init)
            t['properties > post']        .add(properties_post)
            t['stage > init']             .add(stage_init)
            t['docparts > init']          .add(buildWebapps__docparts_init)

            t.execute()
         }
         else
         {
            BuildWebappsTemplate t = new BuildWebappsTemplate(this)

            t['properties > init']        .add(properties_init)
            t['properties > post']        .add(properties_post)
            t['stage > init']             .add(stage_init)
            t['docparts > init']          .add(updateconfigure__docparts_l10n_init)

            t.execute()

         }
      }
      else
      {
         BuildWebappsTemplate t = new BuildWebappsTemplate(this)

         t['properties > init']        .add(properties_init)
         t['properties > post']        .add(properties_post)
         t['stage > init']             .add(stage_init)
         t['docparts > init']          .add(buildWebapps__docparts_init)

         t.execute()
      }
      //api.stop()
   }

   def stage_init =
   {
      String configurableDir              = canonical("$pass._installDir/Configurable")
      String configurableLogDir           = canonical("$pass._installDir/Configurable/LogConfig")
      String commonLogDir                 = api.'config.lev.web.common.log_config.dir'
      String picklistsDir                 = canonical("$pass._installDir/Picklists")
      String picklistFile                 = canonical("$picklistsDir/wars/$pass._rootAppName/picklist")
      String staticDir                    = canonical("$pass._installDir/Static")
      String configurableWarsDir          = canonical("$configurableDir/wars/$pass._rootAppName")
      String configurableEarsDir          = canonical("$configurableDir/ears/$pass._rootAppName")
      String staticWarsDir                = canonical("$staticDir/wars/$pass._rootAppName")
      String staticEarsDir                = canonical("$staticDir/ears/$pass._rootAppName")
      String tempDir                      = api.'config.lev.web.temp.dir'
      String tempEarDir                   = canonical("$tempDir/$pass._rootAppName")
      String tempWarDir                   = canonical("$tempEarDir/$pass._rootAppName")
      String description                  = api.'antimnylndin.description'

      pass.with
      {
         _filters +=
         [
            'webapp.casemgmt.description'          : description,
         ]

         t__stage_extract =
         [
            retrieveFromVJR:
            [
               [
                  packageName: _rootAppName,
                  picklist:    picklistFile,
                  dst:         tempWarDir,
               ],
            ]
         ]

         t__createdirectories =
         [
            prefix: tempDir,
            entries:
            [
               "$_rootAppName/META-INF" as String,
               "$_rootAppName/$_rootAppName/WEB-INF" as String,
            ]
         ]

         t__copyfiles =
         [
            [  src:        configurableEarsDir,
               recurse:    true,
               include:    ~/.*\.orig/,
               dst:        "$tempEarDir/META-INF",
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        configurableWarsDir,
               base:       configurableWarsDir,
               recurse:    true,
               include:    ~/.*\.orig/,
               dst:        tempWarDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        configurableLogDir,
               include:    ~/.*\.orig/,
               dst:        commonLogDir,
               dstfile:    "${_contextName}-log4j.xml",
               strip:      '.orig',
               tokens:     _filters
            ],
         ]

         t__buildarchive =
         [
            buildArchive:
            [
               [
                  appNames  : _rootAppName,
               ]
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value
            :let @/="\\<</key>- <type>String</type>: Property to replace token with</indent></indent>

            \\>"
            </key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def buildWebapps__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
         actions:
            [
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_table_item_URL.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_table_item_logs.txt']
               ],
            ]
         ]

         Boolean inUIP = getProperty('antimnylndin.maintenance.from.version', '0') { toBoolean(it) }
         if (!(t__migrate || inUIP))
         {
            t__docparts.actions +=
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_table_item_validate.txt']
               ]
         }
      }
   }

   // ===================================================== TARGET: deployWebapps

   /**
      Target: deployWebapps

      @see com.sas.gate.api.template.DeployWebappsTemplate
   */
   def deployWebapps()
   {
      //api.stop()

      String maintenanceFrom  = getProperty('antimnylndin.maintenance.from.version', '')

      if (maintenanceFrom)
      {
         String fromVersion         = api.baseVersion(maintenanceFrom)
         String maintenanceTo = getProperty('antimnylndin.maintenance.to.version', '')
         String toVersion           = api.baseVersion(maintenanceTo)
         if (fromVersion != toVersion)
         {
            DeployWebappsTemplate t = new DeployWebappsTemplate(this)

            t.define(
            [
               'properties > pre init',
            ])

            t['properties > pre']         .add(updateconfigure__pre_init)
            t['properties > init']        .add(properties_init)
            t['properties > post']        .add(properties_post)
            t['deploywebapps > init']     .add(deploywebapps_init)

            t.execute()
         }
         else
         {
            DeployWebappsTemplate t = new DeployWebappsTemplate(this)

            t['properties > init']        .add(properties_init)
            t['properties > post']        .add(properties_post)
            t['deploywebapps > init']     .add(deploywebapps_init)

            t.execute()

         }
      }
      else
      {
         DeployWebappsTemplate t = new DeployWebappsTemplate(this)

         t['properties > init']        .add(properties_init)
         t['properties > post']        .add(properties_post)
         t['deploywebapps > init']     .add(deploywebapps_init)

         t.execute()
      }
      //api.stop()
   }

   def deploywebapps_init =
   {
      String stagingDir                   = canonical("${api.'config.lev.web.staging.dir'}/${pass._rootAppName}.ear")

      pass.with {
         t__deploywebapps =
         [
            deploy:
            [
               [
                  pathtoear:  stagingDir,
                  appname:    _applicationName,
               ],
            ]
         ]
      }
   }

   // ===================================================== TARGET: loadContent

   /**
      Target: loadContent

      @see com.sas.gate.api.template.LoadContentTemplate
   */
   def loadContent()
   {
      //api.stop()

      LoadContentTemplate t = new LoadContentTemplate(this)

      t['init']                     .add(upgradeContent_init)
      t['properties > init']        .add(properties_init)
      t['properties > post']        .add(properties_post)

      t['loadcontent > init']       .add(loadcontent_init)
      t['loadattachments > init']   .add(loadattachments_init)

      t['loadworkflows > init']     .add(loadworkflows_init)

      t['createwebservice > init']  .add(createwebservice_init)
      t['docparts > init']          .add(loadContent__docparts_init)

      t.execute()

      //api.stop()
   }

   def upgradeContent_init =
   {
      String maintenanceTo = getProperty('antimnylndin.maintenance.to.version', '')
      if (maintenanceTo)
      {
         componentName     = "${api.'antimnylndin.short.legal.name'} $maintenanceTo"
         appFolder         = "${api.'antimnylndin.system.application.folder'}/${api.'antimnylndin.offering.legal.name'}"
         appSubFolder      = "${appFolder}/${componentName}"
         api.createProperty('antimnylndin.release.number', maintenanceTo)
         api.createProperty('antimnylndin.software.component.name', componentName)
         api.createProperty('antimnylndin.application.folder', appSubFolder)
         api.createProperty('antimnylndin.antimnycmnc.application.folder', "$appFolder/${api.'antimnylndin.offering.short.legal.name'} Svr $maintenanceTo")
      }
   }

   /**
      Callback for hook: <hook>loadcontent &gt; init</hook>

      @param concatenate   Concatenates the custom.properties files from CDD, ECM and AML if included in the plan.
            <indent><key>src</key>- <type>ArrayList&lt;Map&gt;</type>: Array of maps containing the source directories of the files that are to be concatenated<br>
               <indent><key>dir</key>- <type>Map</type>: Map of an array of directories containing files to be concatenated</indent>
            <key>dst</key>- <type>String</type>: Destination where contatinated files will be left including filename, may include regex expression.<br>
            <key>include</key>- <type>String</type>: regular expression used to select files to be concatenated</indent>
      @param menuInject <type>ArrayList&lt;Map&gt;</type>: Inserts the CDD MainNavigationMenu.xml file into the appropriate location in the ECM provided MainNavigationMenu.xml file.
      @param t__loadcontent <type>ArrayList&lt;Map&gt;</type>: Uploads the prepared files to the SAS Content Server.
            <indent><key>failifmissing</key>- <type>Boolean</type>: Fail if files are missing<br>
               <key>content</key>- <type>ArrayList&lt;Map&gt;</type>: Array of content to be uploaded to DAV<br>
            <indent><key>urlSuffix</key>- <type>String</type> <kind>optional</kind>: Suffix to be aplied to the DAV URL where content is to be uploaded to.<br>
            <key>contentLocation</key>- <type>String</type>: Fully qualified path to the Directory where content to be uploaded resides.</indent></indent>
   */
   def loadcontent_init =
   {

      // CONCATENATE: custom*.properties
      String ecmCPDir                     = canonical("${api.'casemgmtmid.install.dir'}/deploy/sample/CustomProperties")
      String fcfInstallConfigSampleDir    = canonical("$pass._installConfigDir/sample")
      String fcfInstallMenuDir            = canonical("$pass._installDir/deploy/menu")
      String fcfInstallUidefDir           = canonical("$pass._installDir/deploy/uidef")
      String configWebRootDir             = canonical("${api.'config.lev.web.applications.dir'}/${api.'amlmid.normalized.offering.name'}${api.'amlmid.release.number'}")
      String amlDirPathCP                 = canonical("$configWebRootDir/install/properties")

      def concatenate =
         [
         src:
            [
               [
                  dir:
                  [
                     ecmCPDir,
                     fcfInstallUidefDir,
                     fcfInstallMenuDir,
                  ]
               ]
            ],
         include: ~/.*?(_.*)*\.properties/,
         overwrite: true,
         dst:     "$pass._configInstallDir/customProperties/custom\\1.properties",
         ]

      new PipeCode(concatenate).concat().write()

      // LOADCONTENT
      String contentPath                  = canonical("$pass._configInstallDir/content")
      String contentsPath                 = canonical("$pass._configInstallDir/Contents")
      String inContentPath                = canonical("$pass._installConfigDir/Deployment/Content")
      String inContentsPath               = canonical("$pass._installConfigDir/Deployment/contents")
      String menuPath                     = canonical("$pass._configInstallDir/menu")
      String fcfDirPathUi                 = canonical("$pass._configInstallDir/uidef")
      String fcfDirPathCP                 = canonical("$pass._configInstallDir/customProperties")
      String fcfWorkflowPath              = canonical("$pass._configInstallDir/workflow")
      String fcfSearchConfigPath          = canonical("$pass._configInstallDir/searchconfig")
      String ecmContentPath               = 'SASEntCaseManagement/EntCaseManagement6.3'
      String amlSuffix                    = '/Products/SASAntiMoneyLaundering'
      String suffix                       = '/Products/SASComplianceSolutions'
      String cddSuffix                    = '/Products/SASCustomerDueDiligence'
      String cpSuffix                     = "/Products/$ecmContentPath/CustomProperties"
      String menuSuffix                   = "/Products/$ecmContentPath/Menus"
      String uiSuffix                     = "/Products/$ecmContentPath/UIDefinition"

      pass.with
      {
         if (api.'wizard.type'=='sdw') t__initializemanualcontent_load = true
         else t__initializemanualcontent_update = true

         t__copyfiles =
         [
            [
               src:        inContentPath,
               include:    ~/.*\.xml/,
               dst:        contentPath,
               tokens:     _filters
            ],
            [
               src:        inContentsPath,
               base:       inContentsPath,
               recurse:    true,
               dst:        contentsPath,
               tokens:     _filters
            ],
            [
               src:        "$pass._installDir/deploy/menu",
               include:    ~/.*\.xml\.orig/,
               dst:        menuPath,
               strip:      '.orig',
               tokens:     _filters
            ],
            [
               src:        "$pass._installDir/deploy/uidef",
               include:    ~/.*\.xml\.orig/,
               dst:        fcfDirPathUi,
               strip:      '.orig',
               tokens:     _filters
            ],
            [
               src:        "$pass._installDir/deploy/workflow",
               include:    ~/.*\.xml/,
               dst:        fcfWorkflowPath,
               strip:      '.orig',
               tokens:     _filters
            ],
            [
               src:        "$pass._installDir/deploy/searchconfig",
               include:    ~/.*\.xml\.orig/,
               dst:        fcfSearchConfigPath,
               strip:      '.orig',
               tokens:     _filters
            ],
         ]

         t__loadcontent_upload =
         [
            failifmissing: true,
            content:
            [
               [
                  contentLocation:  contentPath,
               ],
               [
                  contentLocation:  menuPath,
                  urlSuffix:        menuSuffix,
               ],
               [
                  contentLocation:  contentsPath,
                  urlSuffix:        suffix,
               ],
               [
                  contentLocation:  fcfDirPathUi,
                  urlSuffix:        uiSuffix
               ],
               [
                  contentLocation:  fcfDirPathCP,
                  urlSuffix:        cpSuffix
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>loadattachments &gt; init</hook>

      @param t__loadattachments <type>Map</type>: Uploads the prepared files to the SAS WIP Attachment Service
            <indent><key>password</key>- <type>String</type>: Encoded password generally sasTrust<br>
            <key>urlPrefix</key>- <type>String</type>: Http address of the SASWIPClientAccess/rest/content/<br>
            <key>paths</key>- <type>ArrayList&lt;Map&gt;</type>: Array of maps containing the source that will be uploaded to the attachment service<br>
               <indent><key>dir</key>- <type>String</type>: Fully qualified path to the directory where the files are located that are to be uploaded<br>
               <key>include</key>- <type>String</type>: Regular expression used to select files to be concatenated<br>
               <key>urlSuffix</key>- <type>String</type>: String that will be appended to the urlPrefix to upload the files to the correct location</indent><br>
            </indent>
   */
   def loadattachments_init =
   {
      String password                     = api.'oma.person.trustusr.login.passwd'
      String protocol                     = api.'iomsrv.httpserver.protocol.type'
      String host                         = api.'iomsrv.httpserver.host'
      String port                         = api.'iomsrv.httpserver.port'
      String urlPrefix                    = "$protocol://$host:$port/SASWIPClientAccess/rest/content"
      String ecmUiDir                     = canonical("${api.'casemgmtmid.install.dir'}/deploy/required/UIDefinitions")
      String dirPathUi                    = canonical("$pass._configInstallDir/uidef")
      String dirPathCP                    = canonical("$pass._configInstallDir/customProperties")
      String fcfSearchConfigPath          = canonical("$pass._configInstallDir/searchconfig")

      pass.with
      {
         t__loadattachments =
         [
            password:   password,
            urlPrefix:  urlPrefix,
            paths:
            [
               [dir:          ecmUiDir,
                include:      ~/.*\.xml/,
                urlSuffix:   "545003/UIDefinition/attachments",
               ],
               [dir:          dirPathUi,
                include:      ~/.*\.xml/,
                urlSuffix:   "545003/UIDefinition/attachments",
               ],
               [  dir:          dirPathCP,
                  include:      ~/.*\.properties/,
                  urlSuffix:   "545004/CustomProperties/attachments",
               ],
               [
                  dir:        fcfSearchConfigPath,
                  include:    ~/.*\.xml/,
                  urlSuffix:  "545100/SearchConfig/attachments",
               ],
            ]
         ]
      }
   }

   def loadworkflows_init =
   {
      String workFlowFile              = canonical("$pass._installDir/deploy/workflow/FCFWorkflow.xml")
      pass.with
      {
         t__loadworkflows =
         [  workflows:
            [
               [
                  srcfile: workFlowFile,
                  dst: canonical("$_configInstallDir/workflow"),
               ]
            ]
         ]
      }
   }

   def createwebservice_init =
   {
      String stpPrefix                    = "/System/Applications/$pass._offeringLegalName/$pass._serverSoftwareComponentName/Application SAS Code"
      String picklistsDir                 = canonical("$pass._installDir/Picklists")

      pass.with {
         t__createwebservice =
         [
            services:
            [
               [
                  serviceName       : 'FCFSocialNetworkAnalysis',
                  storedProcesses   :
                  [
                     "$stpPrefix/getSocialNetwork",
                     "$stpPrefix/growSocialNetworkNode",
                     "$stpPrefix/getSocialNetworkNodeDetails"
                  ],
                  namespace         : 'http://sas.com/sso/fraud/sna',
                  publishMetadata   : true,
                  overwrite         : true,
                  picklist          : "$picklistsDir/makeWebService/picklist",
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value
            :let @/="\\<</key>- <type>String</type>: Property to replace token with</indent></indent>

            \\>"
            </key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def loadContent__docparts_init =
   {
      pass.with
      {
         if (!pass.hasVariable('t__docparts')) t__docparts = [:]
         if (!t__docparts.containsKey('actions')) t__docparts.actions = []
         t__docparts.actions +=
         [
            [prepend:[
               token:'instructions.antimnylndin.details.table.item',
               value:api.'instructions_antimnylndin_details_webService.txt']
            ],
            [prepend:[
               token:'instructions.antimnylndin.details.table.item',
               value:api.'instructions_antimnylndin_details_DAVcontent.txt']
            ],
         ]

         if (pass.hasVariable('t__loadworkflows') && pass.t__loadworkflows.docparts)
         {
            pass.t__loadworkflows.docparts.each { workflowFile,uploadSucceeded ->
               String message = uploadSucceeded ? 'instructions_antimnylndin_details_workflow.success.txt'
                  : 'instructions_antimnylndin_details_workflow.failed.txt'

               def value = new PipeCode([api."$message"])
                  .tokenstringsub(tokens:
                  [
                     '_!FILENAME':   workflowFile,
                  ], prefix:'${', suffix:'}').output[0]

               t__docparts.actions += [prepend:[token: 'instructions.antimnylndin.details.table.item', value:value]]
            }
         }
      }
   }
   // ===================================================== TARGET: updateConfigure

   /**
      Target: updateConfigure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def updateConfigure()
   {
      //api.stop()

      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      String maintenanceFrom  = getProperty('antimnylndin.maintenance.from.version', '')

      if (maintenanceFrom)
      {
         String fromVersion         = api.baseVersion(maintenanceFrom)
         String maintenanceTo = getProperty('antimnylndin.maintenance.to.version', '')
         String toVersion           = api.baseVersion(maintenanceTo)
         if (maintenanceFrom != maintenanceTo)
         {
            if (fromVersion != toVersion)
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > pre']            .add(updateconfigure__pre_init)
               t['properties > init']           .add(properties_init)
               t['properties > post']           .add(properties_post)
               t['properties > post']           .add(updateconfigure__properties_post)
               t['properties > post']           .add(properties_assert).when('t__debug')
               t['loadmetadata > init']         .add(updateconfigure__loadmetadata_init)
               t['setobjectproperties > init']  .add(updateconfigure__setobjectproperties_init)
               t['docparts > init']             .add(updateconfigure__docparts_init)

               t.execute()
            }
            else
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > init']        .add(properties_init)
               t['properties > post']        .add(properties_post)
               t['docparts > init']          .add(updateconfigure__docparts_l10n_init)

               t.execute()

            }
         }
      }
      else
         configure()

     //api.stop()
   }

   def updateconfigure__pre_init =
   {
      String legalName = api.createProperty('antimnylndin.short.legal.name', 'Compliance Solutions Mid')

      String toVersion           = api.baseVersion(api.'antimnylndin.maintenance.to.version')
      String fromVersion         = api.baseVersion(api.'antimnylndin.maintenance.from.version')
      String releaseNumber       = toVersion

      String softwareComponent = "$legalName $toVersion"
      String applicationFolder = api.createProperty('antimnylndin.application.folder', "${api.'antimnylndin.system.application.folder'}/${api.'antimnylndin.offering.legal.name'}")
      String applicationSubFolder = api.createProperty('antimnylndin.application.folder', "$applicationFolder/$softwareComponent")

      String configDir= "${api.'config.lev.web.applications.dir'}/${api.'antimnylndin.normalized.offering.name'}"

      api.createProperty('antimnylndin.release.number', toVersion)
      api.createProperty('antimnylndin.software.component.name', softwareComponent)
      api.createProperty('antimnylndin.config.dir', configDir)
      api.createProperty('antimnylndin.configobj.version', api.'antimnylndin.maintenance.to.version')
      api.createProperty('antimnylndin.display.name', softwareComponent)
      api.createProperty('antimnylndin.configobj.name', softwareComponent)
      api.createProperty('antimnylndin.casemgmtmid.software.component.name', 'Ent Case Mgmt Mid-Tier 6.3')
      api.createProperty('antimnylndin.antimnycmnc.application.folder', "$applicationFolder/${api.'antimnylndin.offering.short.legal.name'} Svr $toVersion")
      api.createProperty('antimnylndin.normalized.component.name', "${api.'antimnylndin.normalized.short.legal.name'}${api.'antimnylndin.major.number'}${api.'antimnylndin.minor.number'}")

      pass.with
      {
         _applicationSubFolder = api.createProperty('antimnylndin.application.sub.folder', "$applicationFolder/$softwareComponent")

         switch (fromVersion)
         {
            case(6.1):
               t__removemetadatabypath =
               [
                  paths:
                  [
                     [
                        objectPath     : '/System/Applications/SAS Financial Crimes Foundation',
                        deleteContents : true
                     ],
                  ]
               ]
               break
            default:
               t__removemetadatabypath =
               [
                  paths:
                  [
                     [
                        objectPath     : "$applicationFolder/$legalName $fromVersion",
                        deleteContents : true
                     ],
                  ]
               ]
               break
         }
      }
   }

   def updateconfigure__properties_post =
   {
      pass.with {
         t__createdirectories =
         [
            prefix: _configInstallDir,
            entries:
            [
               'metadata',
               'content',
               'Contents',
               'uidef',
               'menu',
               'customProperties',
               'workflow',
            ]
         ]

         t__copyfiles =
         [
            [  src:        _installMetadataDir,
               include:    ~/.*\.orig/,
               dst:        _configMetadataDir,
               strip:      '.orig',
               tokens:     _filters,
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.locprop/,
               exclude:    ~/.*\.orig/,
               dst:        _configMetadataDir,
            ],
         ]
      }
   }

   def updateconfigure__loadmetadata_init =
   {
      pass.with {
         t__loadmetadata_update =
         [
            file: canonical("$_configMetadataDir/antimnylndin.appxml"),
            folder: _applicationSubFolder,
            applicationname: _softwareComponentName
         ]
      }
   }

   def updateconfigure__setobjectproperties_init =
   {
      pass.with {
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def updateconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  token:'instructions.antimnylndin.head',
                  value:api.'instructions_antimnylndin_head.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.head',
                  value:api.'instructions_antimnylndin_details_table_head.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'updateconfig_antimnylndin_details_table_item_upgrade.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'instructions_antimnylndin_details_doc.txt']
               ],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'updateconfig_antimnylndin_details_table_item_note.txt']
               ],
               [replace:[
                  token:'instructions.antimnylndin.details.table.tail',
                  value:api.'instructions_antimnylndin_details_table_tail.txt']
               ],
            ]
         ]
      }
   }

   def updateconfigure__docparts_l10n_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  token:'instructions.antimnylndin.head',
                  value:api.'instructions_antimnylndin_head.txt']],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.head',
                  value:api.'instructions_antimnylndin_details_table_head.txt']],
               [prepend:[
                  token:'instructions.antimnylndin.details.table.item',
                  value:api.'updateconfig_antimnylndin_details_table_item_l10n.txt']],
               [replace:[
                  token:'instructions.antimnylndin.details.table.tail',
                  value:api.'instructions_antimnylndin_details_table_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: updateContent

   def updateContent()
   {
      //api.stop()
      loadContent()
      //api.stop()
   }

   // ===================================================== TARGET: checkForErrors

   /**
      Target: checkForErrors

      @see com.sas.gate.api.template.CheckForErrorsTemplate
   */
   def checkForErrors()
   {
      CheckForErrorsTemplate t = new CheckForErrorsTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: reportFailures

   /**
      Target: reportFailures

      @see com.sas.gate.api.template.ReportFailuresTemplate
   */
   def reportFailures()
   {
      ReportFailuresTemplate t = new ReportFailuresTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: unconfigure

   /**
      Target: unconfigure

      @see com.sas.gate.api.template.UnconfigureTemplate
   */
   def unconfigure()
   {
      //api.stop()

      UnconfigureTemplate t = new UnconfigureTemplate(this)

      t['properties > init']     .add(properties_init)
      t['properties > post']     .add(properties_post)
      t['init']                  .add(unconfigure__init)
      t['delete > init']         .add(unconfigure__delete_init)
      t['docparts > init']       .add(unconfigure__docparts_init)

      t.execute()

      //api.stop()
   }

   def unconfigure__init =
   {
      def applicationFolder        = api.'antimnylndin.application.folder'

      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : applicationFolder,
                  deleteContents : true
               ],
            ]
         ]

         t__undeployear =
         [
            undeploy:
            [
               [
                  earname: _rootAppName,
                  appname: _applicationName
               ],
            ]
         ]

// TODO: need to research why this doesn't work for CS
//         t__removedatasource =
//         [
//            datasourceName:   _segkcDatasource,
//            servername:       _serverName
//         ]
      }
   }

   def unconfigure__delete_init =
   {
      pass.with
      {
         t__delete_files =
         [
            paths:
            [_configDir]
         ]
      }
   }
   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
            <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def unconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnylndin.head',
                  value:   api.'unconfiguration_antimnylndin_summary_head.txt']
               ],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnylndin.item',
                  value:   api.'unconfiguration_antimnylndin_summary_ear_item.txt']
               ],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnylndin.item',
                  value:   api.'unconfiguration_antimnylndin_summary_sc_item.txt']
               ],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnylndin.item',
                  value:   api.'unconfiguration_antimnylndin_summary_files_item.txt']
               ],
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnylndin.tail',
                  value:   api.'unconfiguration_summary_antimnylndin_tail.txt']
               ],
            ]
         ]
      }
   }

   // ===================================================== TARGET: updatePasswords

   def updatePasswords()
   {
      //api.stop(properties:true)
      UpdatePasswordsTemplate t = new UpdatePasswordsTemplate(this)

      t['properties > init']        .add(properties_init)
      t['properties > post']        .add(properties_post)
      t['updatepasswords > init']   .add(updatepasswords__init)
      t['docparts > init']          .add(updatepasswords__docparts_init)

      t.execute()
      //api.stop(properties:true)
   }

   def updatepasswords__init =
   {

   String dbmsUrl =
      [
         db2:        "jdbc:db2://${api.'antimnylndin.dbms.host'}:${api.'antimnylndin.dbms.port'}/${api.'antimnylndin.dbms.database.name'}",
         oracle:     "jdbc:oracle:thin:@//${api.'antimnylndin.dbms.host'}:${api.'antimnylndin.dbms.port'}/${api.'antimnycmnc.dbms.oracle.service.name'}",
         postgres:   "jdbc:postgresql://${api.'antimnylndin.dbms.host'}:${api.'antimnylndin.dbms.port'}/${api.'antimnylndin.dbms.database.name'}",
         sqlsvr:     "jdbc:sqlserver://${api.'antimnylndin.dbms.host'}\\${api.'antimnylndin.dbms.database.name'}:${api.'antimnylndin.dbms.port'}",
      ]."$pass._dbmsEngine"

     pass.with {
         t__datasource_create =
         [
            adminUser         : api.'antimnylndin.dbms.segkc.userid',
            adminPW           : api.decodeSASstring(api.'antimnylndin.dbms.segkc.passwd'),
            classPath         : _classPath,
            connectionURL     : dbmsUrl,
            datasourceName    : _segkcDatasource,
            driver            : api.'antimnycmnc.dbms.segkc.driver_class',
            jndiName          : _jndiName,
            servername        : _serverName,
            validationQuery   : _validationQuery,
            txType:           'None'
         ]

          t__updatepasswords =
         [
            // Corresponds to the "array" defined by <12byte>.credentials.persisted.list
            // and that for casemgmtmid is defined to be: casemgmtmid.dbms
            // with the implied suffixes: .userid and .passwd
            createdatasource: t__datasource_create
         ]
      }
   }

   def updatepasswords__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            file:    api.'config.lev.updatepasswords.file',
            actions:
            [
               [replace:[
                  token:'updatePasswords.summary.antimnylndin.head',
                  value:api.'updatePasswords_summary_antimnylndin_head.txt']],
               [prepend:[
                  token:'updatePasswords.summary.antimnylndin.item',
                  value:api.'updatePasswords_summary_antimnylndin_kc_item.txt']],
               [replace:[
                  token:'updatePasswords.summary.antimnylndin.tail',
                  value:api.'updatePasswords_summary_antimnylndin_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: changeHost

   def changeHost()
   {
      //api.stop(properties:true)
      ChangeHostTemplate t = new ChangeHostTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['changehost > init']  .add(changehost__init)
      t['docparts > init']    .add(changehost__docparts_init)

      t.execute()
      //api.stop(properties:true)
   }

   def changehost__init =
   {
      Properties props = new Properties()
      File propsFile = new File(api.'configuration.properties.file')
      props.load(propsFile.newDataInputStream())

      pass.with
      {
         t__changehost =
         [
            exact: ['antimnylndin.dbms.host':false],
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def changehost__docparts_init =
   {
      pass.with
      {
         if (t__changehost.docparts)
         {
            t__docparts =
            [
               file:    api.'config.lev.changehost.file',
               actions: []
            ]

            if (t__changehost.docparts.substring)
            {
               def (oldProperties, updatedProperties) = t__changehost.docparts.substring
               oldProperties.each { k, v ->

                  def value = new PipeCode([api.'changeHost_metadata_object_changed.txt'])
                     .tokenstringsub(tokens:
                     [
                        '_!PROPERTY_NAME':   k,
                        '_!OLD_VALUE':       v,
                        '_!NEW_VALUE':       updatedProperties."$k",
                     ], prefix:'${', suffix:'}').output[0]

                  t__docparts.actions += [prepend:[token: 'changeHost.configobj.item', value:value]]
               }
            }
         }
      }
   }
}
