// Copyright 2013-2014, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.

package com.sas.deployment
import groovy.transform.InheritConstructors

import com.sas.gate.CodeAccessor
import com.sas.gate.api.PipeCode
import com.sas.gate.api.template.*

import groovy.io.FileType

/**
   <overview>Deployment code for SAS Compliance Solutions Server (antimnycmnc)</overview>
   <p>
      This code configures the deployment of antimnycmnc using what corresponds to
      the following Ant targets:
      <ul class='nodot'>
         <li>configure</li>
         <li>updateConfigure</li>
         <li>checkForErrors</li>
         <li>reportFailures</li>
         <li>unconfigure</li>
         <li>updatePasswords</li>
         <li>changeHost</li>
      </ul>
   </p>

   @since      14w14
   @version    15w04
   @author     Bruce Bryan
   @author     Steve Amerige
   @see        #SOURCE#
*/

@InheritConstructors
class ANTIMNYCMNC extends CodeAccessor
{
   // ===================================================== GLOBAL

   /**
      Identifier

      @param id <type>String</type>: The 12byte name
   */
   String id = 'ANTIMNYCMNC'

   /**
      Callback for hook: <hook>properties &gt; init</hook>
      <p>This callback is used by every recipie to setup initial properties.</p>

      @param t__properties_get   <type>ArrayList&lt;String&gt;</type>: get object properties on each dependson property

   <ul class='nodot'>
      <li>levconfig</li>
      <li>casemgmtmvac</li>
   </ul>
   */
   def properties_init =
   {
      pass.with {
          // Simple Strings
         _normalizedOfferingLegalName              = api.'antimnycmnc.normalized.offering.legal.name'
         _offeringLegalName                        = api.'antimnycmnc.offering.legal.name'
         _releaseNumber                            = api.'antimnycmnc.release.number'
         _shortLegalName                           = api.'antimnycmnc.short.legal.name'
         _softwareComponentName                    = api.'antimnycmnc.software.component.name'
         _applicationFolder                        = api.'antimnycmnc.system.product.folder'
         _applicationSubFolder                     = api.'antimnycmnc.system.product.version.folder'
         _applicationCodeFolder                    = "$_applicationSubFolder/Application SAS Code"
         _productFolder                            = "/Products/$_offeringLegalName"

         // Config Paths
         _sasConfigDir                             = api.'config.lev.applications.dir'
         _configDir                                = api.'antimnycmnc.config.dir'
         _configInstallDir                         = canonical("$_configDir/install")
         _configMetadataDir                        = canonical("$_configInstallDir/metadata")
         _configCustCfgDir                         = canonical("$_configDir/custom/config")
         _configReportsDir                         = canonical("$_configDir/reports")
         _cddAutoexecPath                          = canonical("$_configCustCfgDir/cdd_autoexec.sas")
         _amlAutoexecPath                          = canonical("$_configCustCfgDir/aml_autoexec.sas")
         _hpamlAutoexecPath                        = canonical("$_sasConfigDir/Applications/SASHighPerformanceAntiMoneyLaundering6.3/custom/config/autoexec.sas")
         _autoexecPath                             = canonical("$_configCustCfgDir/fcf_autoexec.sas")
         _appxmlFile                               = canonical("$_configInstallDir/metadata/antimnycmnc.appxml")

         // Database
         _initDB = !(getProperty('antimnycmnc.bypass_init', '0') { toBoolean(it) })
         _dbmsEngine                               = api.'antimnycmnc.dbms.engine'
         _dbmsDdlDir = 
         [
            db2:        'db2',
            oracle:     'oracle',
            postgres:   'postgres',
            sqlsvr:     'sqlserver',
            teradata:   'teradata',
         ]."$pass._dbmsEngine"
         _configInstallDdlDir                      = canonical("$_configInstallDir/$_dbmsDdlDir")

         t__properties_get =
         [
            'antimnycmnc.dependson.levconfig.fqid',
            'antimnycmnc.dependson.casemgmtmvac.fqid'
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>
      <p>This call back is used by every recipie to setup properties after the depends-on properties
      have been pulled from metadata. These are properties that are common to every recipie.</p>
      <p>Unit testing in this section is performed only if we are in debug mode.</p>

      @param _filters <type>Map&lt;String&gt;</type>: tokens for string substitution
   */
   def properties_post =
   {
      // Simple Strings
      String classIdentifier                    = api.'antimnycmnc.configobj.classid'
      String defaultLocale                      = api.'os.localhost.default.locale'
      String legalName                          = api.'antimnycmn.legal.name'
      String majorNumber                        = api.'antimnycmnc.major.number'
      String minorNumber                        = api.'antimnycmnc.minor.number'
      String offeringShortName                  = api.'antimnycmnc.offering.short.legal.name'

      // Config Paths
      String configCustFmtDir                   = canonical("$pass._configDir/custom/fmts")
      String configStageDir                     = canonical("$pass._configDir/data/stage")
      String configIemDir                       = canonical("$pass._configDir/data/iem")
      String configJobstatsDir                  = canonical("$pass._configDir/job_stats")
      String configReportsDir                   = canonical("$pass._configDir/reports")
      String configLandingDir                   = canonical("$pass._configDir/landing")
      String ecmAutoexecPath                    = canonical("${api.'config.lev.applications.dir'}/SASEnterpriseCaseManagement/Source/control/ecm_autoexec.sas")
      String ecmInstallMiscDir = (api.'os.localhost.type' == 'win') ? canonical("${api.'sasinstall.install.dir'}/casemgmtmva/sasmisc") : canonical("${api.'sasinstall.install.dir'}/misc/casemgmtmva")
      String installSASExecFile                 = api.'sas.exec.file'

      // Install Paths (Dirs and Files)
      String installSASDir                      = api.'sasinstall.install.dir'
      String installSASConfigFile               = canonical("$installSASDir/sasv9.cfg")
      String installCmacrosDir                  = api.'antimnycmnc.install.cmacros.dir'
      String installUcmacrosDir                 = api.'antimnycmnc.install.ucmacros.dir'
      String sasConfigFile                      = api.'sas.config.file'

      // Other
      String serverContextName                  = api.'oma.appserver.servercontext.name'
      String authdomain                         = api.'iomsrv.metadatasrv.authdomain'
      String businessUnitName                   = api.'antimnycmnc.business.unit.name'
		String amlGroupName                       = 'AML Scenario Administrators'
      String coreUserId                         = api.'antimnycmnc.dbms.core.userid'
      String kcUserId                           = api.'antimnycmnc.dbms.kc.userid'
      String segkcUserId                        = api.'antimnycmnc.dbms.segkc.userid'
      String segmentId                          = segkcUserId
      String ecmDbEngine                        = api.'casemgmtmvac.dbms.engine'
      String ecmDbConopts                       = [
                                                    db2       : 'database=' + api.'casemgmtmvac.dbms.database.name',
                                                    oracle    : 'path=' + api.'casemgmtmvac.dbms.oracle.sid',
                                                    sqlsvr    : 'dsn=' + api.'casemgmtmvac.dbms.database.name',
                                                    odbc      : 'dsn=' + api.'casemgmtmvac.dbms.database.name',
                                                    postgres  : 'server=\"' + api.'casemgmtmvac.dbms.host' + '\"' + ' database=' + api.'casemgmtmvac.dbms.database.name' + ' port=' + api.'casemgmtmvac.dbms.port',
                                                   ]."$ecmDbEngine"

      String installUser                        = api.'os.localhost.userid'
      String scenarioMessageFile                = canonical("$pass._configDir/scenario/scenario_source/fsk_scenario_message.csv")

      // Properties
      api.createProperty('antimnycmnc.autoexec.path', pass._autoexecPath)
      api.createProperty('antimnycmnc.aml.autoexec.path', pass._amlAutoexecPath)
      String davPath = getProperty('antimnycmnc.dav.path', "Products/$pass._normalizedOfferingLegalName")
      api.createProperty('antimnycmnc.dav.path', davPath)
      api.createProperty('antimnycmnc.configobj.name', pass._softwareComponentName)

      if (t__debug){
         assert classIdentifier == 'antimnycmnc' : "Class identifier does not match: antimnycmnc"
         assert legalName == 'SAS Compliance Solutions Server' : "Legal name does not match: SAS Compliance Solutions Server"
         assert majorNumber == '6' : "Release major number does not match: 6"
         assert minorNumber == '3' : "Release minor number does not match: 3"
         assert offeringShortName == 'Compliance Solutions' : "Offering short legal name does not match: Compliance Solutions"
      }

      pass.with
      {
         _installConfigDir                   = api.'antimnycmnc.install.config.dir'
         _installPackagesDir                 = canonical("$_installConfigDir/Deployment/Packages")
         _installSampleDir                   = api.'antimnycmnc.install.sample.dir'
         _installSasmiscDir                  = api.'antimnycmnc.install.sasmisc.dir'
         _installDeployDir                   = canonical("$_installSasmiscDir/deploy")
         _installReportsDir                  = canonical("$_installSasmiscDir/reports")
         _installDdlDir                      = canonical("$_installSasmiscDir/ddl/$_dbmsDdlDir")
         _installStpDir                      = api.'antimnycmnc.install.sasstp.dir'
         _configCustsrcDir                   = canonical("$_configDir/custom/source")
         _configDataDir                      = canonical("$_configDir/data")
         _configStpDir                       = canonical("$_configDir/sasstp")
         _coreGroup                          = 'Core'
         _kcGroup                            = 'Knowledge Center'
         _segkcGroup                         = 'Business Unit Knowledge Center'
         _installSASNLSConfigFile            = canonical("$installSASDir/nls/u8/sasv9.cfg")
         _installUnxU8sasexec                = canonical("$installSASDir/bin/sas_u8")
         _dormantDbPrep                      = "dormant_account_${_dbmsEngine}_prep.sas"
         _segkcAuth                          = api.'antimnycmnc.dbms.segkc.auth'
         _segkcConnOpts                      = api.'antimnycmnc.dbms.segkc.connopts'
         _coreAuth                           = getProperty('antimnycmnc.dbms.core.auth', 'Core')
         _coreConnOpts                       = getProperty('antimnycmnc.dbms.core.connopts', _segkcConnOpts)
         _kcAuth                             = getProperty('antimnycmnc.dbms.kc.auth', 'Knowledge Center')
         _kcConnOpts                         = getProperty('antimnycmnc.dbms.kc.connopts', _segkcConnOpts)
         _coreUserId                         = api.'antimnycmnc.dbms.core.userid'
         _kcUserId                           = api.'antimnycmnc.dbms.kc.userid'
         _segkcUserId                        = api.'antimnycmnc.dbms.segkc.userid'
         _dbmsDdlTmpDir                      = canonical("$_configDir/tmp/$_dbmsDdlDir")
         _dbmsHost                           = api.'antimnycmnc.dbms.host'
         _dbmsDdlList =
         [
            db2       : [],
            oracle    :
            [
               [ file:'oracle_core_ddl.sql',               authdomain:_coreAuth, connopts:_coreConnOpts  ],
               [ file:'oracle_knowledge_center_ddl.sql',   authdomain:_kcAuth,   connopts:_kcConnOpts    ],
               [ file:'oracle_grants_core.sql',            authdomain:_coreAuth, connopts:_coreConnOpts  ],
               [ file:'oracle_grants_kc.sql',              authdomain:_kcAuth,   connopts:_kcConnOpts    ],
               [ file:'oracle_bu1_ddl.sql',                authdomain:_segkcAuth,connopts:_segkcConnOpts ],
            ],
            sqlsvr    : [],
            odbc      : [],
            postgres  : [],
         ]."$_dbmsEngine".collect { Map m -> "$m.file|$m.authdomain|$m.connopts" }.join('/')

         println "##### _dbmsDdlList = $_dbmsDdlList"

         _filters =
         [
            'agp.autoroute'                        : api.'antimnycmnc.agp.autoroute',
            'agp.mlrisk'                           : api.'antimnycmnc.agp.mlrisk',
            'agp.replicate.suppressed.alerts'      : api.'antimnycmnc.agp.replicate.suppressed.alerts',
            'agp.scoring.duration'                 : api.'antimnycmnc.agp.scoring.duration',
            'agp.tfrisk'                           : api.'antimnycmnc.agp.tfrisk',
            'aml.autoexec.path'                    : _amlAutoexecPath,
            'aml.custfmt'                          : configCustFmtDir,
            'aml.custsrc'                          : _configCustsrcDir,
            'aml.sa.group.name'                    : amlGroupName,
            'aml.group.name'                       : _segkcGroup,
            'amldata'                              : _configDataDir,
            'amlroot'                              : _configDir,
            'appserver.servercontext.name'         : serverContextName,
            'authdomain'                           : authdomain,
            'cdd.autoexec.path'                    : _cddAutoexecPath,
            'cdd.custfmt'                          : configCustFmtDir,
            'cdd.custsrc'                          : _configCustsrcDir,
            'cdd.install.ucmacros.dir'             : installUcmacrosDir,
            'cdddata'                              : _configDataDir,
            'cddroot'                              : _configDir,
            'class.identifier'                     : classIdentifier,
            'cmacros.dir'                          : installCmacrosDir,
            'cmndata'                              : _configDataDir,
            'cmnroot'                              : _configDir,
            'custfmt'                              : configCustFmtDir,
            'custsrc'                              : _configCustsrcDir,
            'data.stage'                           : configStageDir.replaceAll('\\\\', '\\\\\\\\'),
            'dav.path'                             : davPath,
            'aml.dav.path'                         : davPath,  // put these two back in temporarily
            'cdd.dav.path'                         : davPath,  // put these two back in temporarily
            'dbms.core.connopts'                   : _coreConnOpts,
            'dbms.core.passwd'                     : api.'antimnycmnc.dbms.core.passwd',
            'dbms.core.schema'                     : api.'antimnycmnc.dbms.core.schema',
            'dbms.core.userid'                     : _coreUserId,
            'dbms.ddl.list'                        : _dbmsDdlList,
            'dbms.ddl.temp.dir'                    : _dbmsDdlTmpDir,
            'dbms.engine'                          : _dbmsEngine,
            'dbms.host'                            : _dbmsHost,
            'dbms.kc.connopts'                     : _kcConnOpts,
            'dbms.kc.passwd'                       : api.'antimnycmnc.dbms.kc.passwd',
            'dbms.kc.schema'                       : api.'antimnycmnc.dbms.kc.schema',
            'dbms.kc.userid'                       : _kcUserId,
            'dbms.libopts'                         : api.'antimnycmnc.dbms.libopts',
            'dbms.mode'                            : api.'antimnycmnc.dbms.mode',
            'dbms.segkc.connopts'                  : _segkcConnOpts,
            'dbms.segkc.passwd'                    : api.'antimnycmnc.dbms.segkc.passwd',
            'dbms.segkc.schema'                    : api.'antimnycmnc.dbms.segkc.schema',
            'dbms.segkc.userid'                    : _segkcUserId,
            'dbUser'                               : _kcUserId,
            'default.locale'                       : defaultLocale,
            'description'                          : _softwareComponentName,
            'ecm.autoexec.path'                    : ecmAutoexecPath,
            'ecm.db.conopts'                       : ecmDbConopts,
            'ecm.misc.dir'                         : ecmInstallMiscDir,
            'email.footer'                         : canonical("$_configCustsrcDir/email_footer.txt"),
            'fcf.autoexec.path'                    : _autoexecPath,
            'fcf.config.ddl.path'                  : _configInstallDdlDir,
            'fcf.install.ucmacros.dir'             : installUcmacrosDir,
            'fcf.legal.short.name'                 : offeringShortName,
            'fcf.software.component.name'          : api.'antimnycmnc.antimnylndin.software.component.name',
            'fcf.stp'                              : _configStpDir,
            'fcfdata'                              : _configDataDir,
            'fcfland'                              : configLandingDir,
            'fcfroot'                              : _configDir,
            'full.core.lib'                        : api.'antimnycmnc.dbms.core.library.name',
            'full.kc.lib'                          : api.'antimnycmnc.dbms.kc.library.name',
            'hpaml.autoexec.path'                  : _hpamlAutoexecPath,
            'iem.path'                             : configIemDir,
            'install.cmacros.dir'                  : installCmacrosDir,
            'install.misc.dir'                     : _installSasmiscDir,
            'install.xml.map'                      : canonical("$_installSasmiscDir/xml/map"),
            'installuser'                          : installUser,
            'iomsrv.metadatasrv.host'              : api.'iomsrv.metadatasrv.host',
            'iomsrv.metadatasrv.port'              : api.'iomsrv.metadatasrv.port',
            'jobstat'                              : configJobstatsDir,
            'major.number'                         : majorNumber,
            'minor.number'                         : minorNumber,
            'month.length'                         : api.'antimnycmnc.agp.month.length.num',
            'nefs'                                 : api.'file.separator',
            'normalized.offering.name'             : _normalizedOfferingLegalName,
            'oma.repository.foundation.name'       : api.'oma.repository.foundation.name',
            'package.dir'                          : _installPackagesDir,
            'product.description'                  : _softwareComponentName,
            'product.folder'                       : _applicationSubFolder,
            'product.name'                         : _shortLegalName,
            'product.short.name'                   : _softwareComponentName,
            'product.version.folder'               : _applicationSubFolder,
            'release.number'                       : _releaseNumber,
            'sas.config.file'                      : sasConfigFile,
            'sas.exec.file'                        : installSASExecFile,
            'scenario.message'                     : scenarioMessageFile,
            'scenario.type'                        : api.'antimnycmnc.scenario.kind',
            'segment.settings'                     : canonical("$_configDir/install/segment_settings.sas"),
            'segment_id'                           : segmentId,
            'software.component.name'              : _softwareComponentName,
            'source_system_cd'                     : api.'antimnycmnc.dbms.segkc.userid',
            'stp.location'                         : _applicationCodeFolder,
            'stproot'                              : _configStpDir.replaceAll('\\\\', '\\\\\\\\'),
            'sysadmin.email'                       : api.'antimnycmnc.sysadmin.email',
         ]

      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>
      <p>This is a Unit testing callback that is performed only if we are in debug mode.</p>
   */
   def properties_assert =
   {
      pass.with {
         assert _offeringLegalName == 'SAS Compliance Solutions' : "Offering legal name does not match: SAS Compliance Solutions"
         assert _releaseNumber == '6.3' : "SAS Compliance Solutions release number does not match: 6.3"
         assert _shortLegalName == 'Compliance Solutions Svr' : "Short legal name does not match: Compliance Solutions Svr"
         assert _softwareComponentName == 'Compliance Solutions Svr 6.3' : "Software component name does not match: Compliance Solutions Svr 6.3"
         assert _applicationFolder == '/System/Applications/SAS Compliance Solutions' : "Application metadata folder does not match: /System/Applications/SAS Compliance Solutions"
         assert _applicationSubFolder == '/System/Applications/SAS Compliance Solutions/Compliance Solutions Svr 6.3' : "Application metadata sub folder does not match: /System/Applications/SAS Compliance Solutions/Compliance Solutions Svr 6.3"
         assert _applicationCodeFolder == '/System/Applications/SAS Compliance Solutions/Compliance Solutions Svr 6.3/Application SAS Code' : "Application metadata sub folder does not match: /System/Applications/SAS Compliance Solutions/Compliance Solutions Svr 6.3/Application SAS Code"
         assert _productFolder == '/Products/SAS Compliance Solutions' : "Product metadata folder does not match: /Products/SAS Compliance Solutions"
         assert _configDir =~ "SASComplianceSolutions" : "Config dir does not contain: SASComplianceSolutions"
         assert _configInstallDir =~ "install" : "Config dir does not contain: install"
         assert _configMetadataDir =~ "metadata" : "Config dir does not contain: metadata"
         assert _configCustCfgDir =~ "custom" : "Config dir does not contain: custom"
         assert _configCustCfgDir =~ "config" : "Config dir does not contain: config"
      }
   }

   // ===================================================== TARGET: configure

   /**
      Target: configure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def configure()
   {
      //api.stop()
      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      ConfigureTemplate t = new ConfigureTemplate(this)
      t.dumptreeinfo()

      t['migration > init']         .add(configure__migration_init)
      t['properties > init']        .add(properties_init)
      t['properties > post']        .add(properties_post)
      t['properties > post']        .add(configure__properties_post)
      t['properties > post']        .add(configure__migrate_properties_post).when('t__migrate')
      t['properties > post']        .add(properties_assert).when('t__debug')
      t['updatemetadata > init']    .add(configure__updatemetadata_init).when('t__migrate')

      t['loadmetadata > migration > init'] .add(configure__loadmetadata_migration_init).when('t__migrate')
      t['docparts > init']          .add(configure__docparts_init).when('!t__migrate')
      t['docparts > migration']     .add(configure__docparts_migration).when('t__migrate')
      t['updatepasswords > init']   .add(updatepasswords__init).when('t__migrate')

      t['sasjobs > createuser']     .add(sasjobs_createuser).when('!t__migrate || _upgrading')
      t['sasjobs > removeuser']     .add(sasjobs_removeuser).when('!t__migrate || _upgrading')

      t['permissions']              .add(configure__permissions).when("__os == 'unx'")

      t.execute()

      //api.stop()
   }

   def configure__permissions =
   {
      CodeAccessor a = pass.__template.accessor.createCodeAccessor('com.sas.gate.api.ChmodCode')
      a.api.chmod([paths:pass._configDir, recurse:true, file:'ug=rw,o=r', dir:'ug=rwx,o=rx'])
   }

   /**
      Callback for hook: <hook>migration &gt; init</hook>
      <p>This call back is used when in migration and sets up various properties that will be needed
      for the migration to work, it also contains business logic for determinition of what is required
      to be done depending upon the version that is being migrated from.</p>

      @param t__migration_searchandcall <type>Map&lt;ArrayList&gt;</type>: contains actions that will set
      a metadata type and a search string to be inserted into an XMLSelect statement.
      @param t__searchandcall <type>Map&lt;ArrayList&gt;</type>: contains actions that will be
      performed a metadata type and a search string to be inserted into an XMLSelect statement.
      @param updateconfigure__loadmetadata_init <type>Closure</type>: Sets up properties for running
      updateApplicationMetadata instead of createApplicationMetadata
   */
   def configure__migration_init =
   {
      String releaseNumber = api.'antimnycmnc.release.number'
      String configDir = canonical("${api.'config.lev.applications.dir'}/${api.'antimnycmnc.normalized.offering.legal.name'}")
      String autoexecPath = canonical("${configDir}/custom/config/fcf_autoexec.sas")
      api.createProperty('antimnycmnc.config.dir', configDir)
      api.createProperty('antimnycmnc.autoexec.path', autoexecPath)
      pass.with {
	      _upgrading = false
         _maintenanceFrom = getProperty('antimnycmnc.maintenance.from.version', '')
         t__migrate = getProperty('antimnycmnc.migration.is_enabled', null)
         if (t__migrate == null)
         {
            Boolean migrationSelected = getProperty('config.migration.is_selected', '0') { toBoolean(it) }
            String commonFile = "${api.'config.migration.package.tier.dir'}/antimnylndcf/common.properties"

            t__migrate = migrationSelected && (new File(commonFile).exists())
         }
         else
            t__migrate = toBoolean(t__migrate)

         _updateMetadata = false

         if (t__migrate)
         {
            _migrationFrom = getProperty('antimnycmnc.migration.from.group', '_51_and_earlier')
            switch (_migrationFrom)
            {
            case '_51_and_earlier':
               t__removemetadatabypath =
               [
                  paths:
                  [
                     [
                        deleteContents : true,
                        objectPath : "/System/Applications/SAS Anti-Money Laundering"
                     ]
                  ]
               ]
               t__migrate = false
               break

            case '_52_and_61':
		         _upgrading = true

               t__removemetadatabypath =
               [
                  paths:
                  [
                     [
                        deleteContents : true,
                        objectPath : "/System/Applications/SAS Anti-Money Laundering"
                     ]
                  ]
               ]

               if (_maintenanceFrom == '6.1')
               {
                  t__removemetadatabypath =
                  [
                     paths:
                     [
                        [
                           deleteContents : true,
                           objectPath : "/System/Applications/SAS Anti-Money Laundering"
                        ],
                        [
                           deleteContents : true,
                           objectPath : "/System/Applications/SAS Customer Due Diligence"
                        ],
                        [
                           deleteContents : true,
                           objectPath : "/System/Applications/SAS Financial Crimes Foundation/Financial Crimes Fdn $_maintenanceFrom/Application SAS Code"
                        ]
                     ]
                  ]

                  t__searchandcall =
                  [
                     actions: [ __template.t__deletemetadata ],
                     fqids :  [
                                 [type:'SASLibrary', search:"@Name contains 'FCF -'"],
                                 [type:'SASClientConnection', search:"@Name contains 'Core' or @Name contains 'Knowledge Center'"],
                                 [type:'ServerComponent', search:  "@Name contains 'FCF -'"]
                              ]
                  ]
               }
               break

            case '_62_and_later':
               _updateMetadata = true
               updateconfigure__loadmetadata_init()
               if ( releaseNumber != _maintenanceFrom )
               {
		            _upgrading = true

                  t__removemetadatabypath =
                  [
                     paths:
                     [
                        [
                           deleteContents : true,
                           objectPath : "/System/Applications/${api.'antimnycmnc.offering.legal.name'}/${api.'antimnycmnc.short.legal.name'} ${api.'antimnycmnc.maintenance.from.version'}"
                        ]
                     ]
                  ]
               }
               break

            default:
               break
            }
         }
      }
   }

   /**
      Callback for hook: <hook>updatemetadata &gt; init</hook>
      <p>This callback is used when in migration and sets up a property that will be used to rename
      the Business Unit Knowledge Center IdentityGroup in metadata.</p>

      @param t__updatemetadata <type>Map&lt;ArrayList&gt;</type>: contains actions that will be
      performed a metadata type and a search string to be inserted into an XMLSelect statement.
   */
   def configure__updatemetadata_init =
   {
      String oldSegkcUserId = api.'antimnycmnc.dbms.segkc.userid.orig'

      pass.with {
      switch (_maintenanceFrom)
      {
      case '6.2':
         break
      case '6.1':
            if (oldSegkcUserId != _segkcUserId)
            {
            t__updatemetadata =
            [
               updates:
               [
                  [
                     type:    'IdentityGroup',
                     search:  "@Name='$oldSegkcUserId Knowledge Center'",
                     newName: 'Business Unit Knowledge Center'
                  ]
               ]
            ]
         }
            else
            {
               t__updatemetadata =
               [
                  updates:
                  [
                     [
                        type:    'IdentityGroup',
                        search:  "@Name='$_segkcUserId Knowledge Center'",
                        newName: 'Business Unit Knowledge Center'
                     ]
                  ]
               ]
            }
         break
      default:
         break
      }
   }
   }

   /**
      Callback for hook: <hook>loadmetadata &gt; migration &gt; init</hook>
      <p>This call back is used when in migration and sets up a property that will be used to rename
      the Business Unit Knowledge Center IdentityGroup in metadata.</p>

      @param t__loadmetadata_migration_createfolders <type>ArrayList&lt;String&gt;</type>: contains a list of folders
      tp be created in metadata.
      @param t__loadmetadata_migration_update <type>Map&lt;String&gt;</type>: Properties to be used by the AMU in update mode.
      @param t__importpackages_migration_run <type>Map&lt;ArrayList&gt;</type>: SAS packages that will need to be imported during migration.
   */
   def configure__loadmetadata_migration_init =
   {
      pass.with {
         switch (_migrationFrom)
         {
         case '_62_and_later':
            if ( _releaseNumber != _maintenanceFrom )
            {
               t__loadmetadata_migration_createfolders =
               [
                  _applicationSubFolder,
                  _applicationCodeFolder
               ]
            }
            break

         default:
            t__loadmetadata_migration_createfolders =
            [
               _productFolder,
               _applicationFolder,
               _applicationSubFolder,
               _applicationCodeFolder
            ]
            break
         }

         t__loadmetadata_migration_update =
         [
            file: _appxmlFile,
            folder: _applicationSubFolder,
            applicationname: _softwareComponentName
         ]

         t__importpackages_migration_run =
         [
            packages:
            [
               [
               name:          'ComplianceSolutions',
               packagesDir:   canonical("$_installConfigDir/Deployment/Packages/"),
               folder:        _applicationCodeFolder,
               tokens:        _filters
               ]
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>
      <p>This call back is used to set up properties that will be nneeded in configuration.</p>

      @param t__createdirectories <type>Map&lt;String&gt;</type>: directories to be created in configure target<br>
         <indent><key>prefix</key>- <type>String</type>: base directory to create<br>
         <key>entries</key>- <type>ArrayList&lt;String&gt;</type>: directories and subdirectories to be created<br></indent>
      <pre>
config/Lev1/Applications/SASComplianceSolutions
|-- backups
|-- classifier
|   |-- ca
|   `-- codegen
|-- custom
|   |-- config
|   |-- fmts
|   `-- source
|       `-- prep_files
|-- data
|   |-- iem
|   |-- master
|   |   |-- alert
|   |   |-- bridge
|   |   |-- ca
|   |   |-- crr
|   |   |-- dim
|   |   |-- fact
|   |   |-- nn
|   |   |-- prep
|   |   |-- rc
|   |   |-- report
|   |   |-- sar
|   |   `-- watchlist
|   |-- rule
|   |-- scenario
|   |   |-- scenario_source
|   |   |-- scenario_test
|   |   `-- xml
|   `-- stage
|       |-- alert
|       |-- bridge
|       |-- ca
|       |-- control
|       |-- dim
|       |-- error
|       |-- fact
|       |-- rc
|       |   |-- acc
|       |   |-- ppf
|       |   `-- pty
|       |-- score
|       |-- watchlist
|       `-- xref
|-- fcfsync
|-- install
|   |-- data
|   |-- metadata
|   `-- xml
|-- job_stats
|-- landing
|-- logs
|-- reports
|   `-- xml
|-- rule
|   |-- codegen
|   |-- rule_code_active
|   `-- rule_group_code_active
|-- sasstp
|-- sasuser
`-- scenario
    |-- codegen
    |-- header_code_active
    |-- header_source
    |-- scenario_code_active
    |-- scenario_source
    `-- scenario_test</pre>

      @param t__setobjectproperties    <type>ArrayList&lt;String&gt;</type>: Properties that will be stored in the configured objects text store.
      @param t__copyfiles  <type>ArrayList&lt;Map&gt;</type>: files to copy and conditions for the copy<br>
            <indent><key>src</key>- <type>String</type>: Source directory to copy from<br>
            <key>include</key>- <type>String</type>: Regex pattern of file(s) to copy<br>
            <key>dst</key>- <type>String</type>: Destination directory for files to be copied to<br>
            <key>strip</key>- <type>String</type> <kind>optional</kind>: Remove extension form file name i.e., remove .orig extension<br>
            <key>tokens</key>- <type>String</type> <kind>optional</kind>: filterset to use for token replacement<br></indent>
      @param t__loadmetadata_createfolders <type>ArrayList&lt;String&gt;</type>: Contains a list of folders that will be created in metadata during configuration.
      @param t__loadmetadata_load   <type>String</type>: Fully qualified path to the token substituted appxml file
      @param t__importpackages_createfolders <type>ArrayList&lt;String&gt;</type>: Contains folder(s) that will be created in metadata during configuration.
      @param t__importpackages_run  <type>Map&lt;Arraylist&gt;</type>: Contains properties required to import package(s) into metadata.<br>
            <indent><key>packages</key>- <type>Arraylist&lt;Map&gt;</type>: SAS packages to be imported<br>
               <indent><key>name</key>- <type>String</type>: name of the package file without extension<br>
               <key>packagesDir</key>- <type>String</type>: Fully wualified path to directory containing the package to be imported<br>
               <key>folder</key>- <type>String</type>: Folder in metadata that will hold the imported package<br>
               <key>tokens</key>- <type>Map</type>: used to token substitute the subprop files of the package<br></indent></indent>
      @param t__addlogintogroup  <type>Map&lt;Arraylist&gt;</type>: Authdomain(s) and user to be associated with each authdomain.
      @param t__sasjobs_run  <type>Map&lt;Arraylist&gt;</type>:
            <indent><key>enabled</key>- <type>Bool</type> <kind>optional</kind>: Is the datastructure enabled.  Default value: true<br>
            <key>folder</key>- <type>String</type>: directory on disk where code to be run exists<br>
            <key>groups</key>- <type>ArrayList</type>: Groups containing authentication domain credentials in metadata<br>
            <key>jobs</key>- <type>ArrayList</type>: Remove extension form file name i.e., remove .orig extension<br>
               <indent><key>jobname</key>- <type>String</type>: name of the file on disk not including the extension<br>
               <key>win</key>- <type>Map</type>: Windows specific options for running a SAS job.<br>
               <key>unx</key>- <type>Map</type>: Unix specific options for running a SAS job.<br>
               <key>message</key>- <type>String</type>: MEssage to be displyed in Warnings section of Instructions.html file<br>
            </indent></indent>
   */
   def configure__properties_post =
   {
      // SAS Jobs
      String sasFCFInsJob                       = "fcf_install"
      String sasFCFECMJob                       = "fcf_ecm_config"
      String sasAMLInsJob                       = "aml_install"
      String sasAMLECMJob                       = "aml_ecm_config"
      String sasCDDInsJob                       = "cdd_install"
      String sasCDDECMJob                       = "cdd_ecm_config"
      String sasJobFCFInsFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasFCFInsJob],prefix:'${',suffix:'}').output[0]
      String sasJobFCFECMFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasFCFECMJob],prefix:'${',suffix:'}').output[0]
      String sasJobAMLInsFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasAMLInsJob],prefix:'${',suffix:'}').output[0]
      String sasJobAMLECMFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasAMLECMJob],prefix:'${',suffix:'}').output[0]
      String sasJobCDDInsFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasCDDInsJob],prefix:'${',suffix:'}').output[0]
      String sasJobCDDECMFailed                 = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasCDDECMJob],prefix:'${',suffix:'}').output[0]

      pass.with
      {
         t__createdirectories =
         [
            prefix: _configDir,
            chmod:  'ug=rwx,o=rx',
            recurse: true,
            entries:
            [
               'backups',
               'classifier/ca',
               'classifier/codegen',
               'custom/config',
               'custom/fmts',
               'custom/source',
               'custom/source/prep_files',
               'data/iem',
               'data/master/alert',
               'data/master/bridge',
               'data/master/ca',
               'data/master/crr',
               'data/master/dim',
               'data/master/fact',
               'data/master/nn',
               'data/master/prep/cdd',
               'data/master/rc',
               'data/master/report',
               'data/master/sar',
               'data/master/watchlist',
               'data/rule',
               'data/scenario/scenario_source',
               'data/scenario/scenario_test',
               'data/scenario/xml',
               'data/solr',
               'data/stage/alert',
               'data/stage/bridge',
               'data/stage/ca',
               'data/stage/control',
               'data/stage/dim',
               'data/stage/error',
               'data/stage/fact',
               'data/stage/rc/acc',
               'data/stage/rc/ppf',
               'data/stage/rc/pty',
               'data/stage/score',
               'data/stage/watchlist',
               'data/stage/xref',
               'fcfsync',
               'job_stats',
               'install/data',
               'install/metadata',
               "install/$_dbmsDdlDir" as String,
               'landing',
               'logs',
               'reports/xml',
               'rule/codegen',
               'rule/rule_code_active',
               'rule/rule_group_code_active',
               'sasuser',
               'sasstp',
               'scenario/codegen',
               'scenario/header_code_active',
               'scenario/header_source',
               'scenario/scenario_source',
               'scenario/scenario_code_active',
               'scenario/scenario_test',
               "tmp/$_dbmsDdlDir" as String,
            ]
         ]

         t__delete_files = [paths:["$_configDir/tmp"]]

         t__setobjectproperties =
         [
            properties:
            [
               'antimnycmnc.bypass_init',
               'antimnycmnc.dav.path',
               'antimnycmnc.release.number',
               'antimnycmnc.software.component.name',
            ]
         ]

         t__setobjectproperties.properties +=
         (_dbmsEngine == 'oracle') ?
            [
               'antimnycmnc.dbms.oracle.sid'
            ]
         :
            [
               'antimnycmnc.dbms.database.name'
            ]

         t__copyfiles =
         [
            [  src:        _installSasmiscDir,
               include:    ~/.*\.sas/,
               dst:        _configCustsrcDir
            ],
            [  src:        _installSampleDir,
               include:    ~/.*\.(sas|cmd|csv|sh)/,
               exclude:    ~/.*(_prep\.sas|\.sas7bdat)/,
               dst:        _configCustsrcDir,
               tokens:     _filters
            ],
            [  src:        _installSampleDir,
               include:    ~/(fcf_prep|fcf_iem_.*)\.sas/,
               dst:        _configCustsrcDir,
               tokens:     _filters
            ],
            [  src:        _installSampleDir,
               include:    ~/.*_prep\.sas/,
               exclude:    ~/(fcf_prep|.*db2.*|.*mysql.*|.*oracle.*|.*teradata.*|.*odbc.*|.*sqlsvr.*|.*postgres.*)\.sas/,
               dst:        "$_configCustsrcDir/prep_files",
               tokens:     _filters
            ],
            [  src:        _installSampleDir,
               include:    _dormantDbPrep,
               dst:        "$_configCustsrcDir/prep_files",
               tokens:     _filters
            ],
            [  src:        _installSampleDir,
               include:    ~/.*\.sas7bdat/,
               dst:        _configDataDir
            ],
            [  src:        _installStpDir,
               include:    ~/.*\.(orig|sas)/,
               dst:        _configStpDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installDeployDir,
               include:    ~/.*autoexec\.sas/,
               dst:        _configCustCfgDir,
               tokens:     _filters
            ],
            [  src:        _installDeployDir,
               include:    ~/.*(install|config|sa|update.*)\.sas/,
               dst:        _configInstallDir,
               tokens:     _filters
            ],
            [  src:        _installDdlDir,
               include:    ~/.*\.sql/,
               dst:        _configInstallDdlDir,
               tokens:     _filters
            ],
            [  src:        _installReportsDir,
               include:    ~/.*\.sas/,
               dst:        _configReportsDir
            ],
            [  src:        "$_installDeployDir/data",
               include:    ~/.*/,
               exclude:    ~/fsk_scenario_message\.csv/,
               dst:        "$_configInstallDir/data",
               tokens:     _filters
            ],
            [  src:        "$_installDeployDir/data",
               include:    ~/fsk_scenario_message\.csv/,
               dst:        "$_configDir/scenario/scenario_source",
               tokens:     _filters
            ],
            [  src:        "$_installDeployDir/xml",
               include:    ~/.*/,
               dst:        "$_configInstallDir",
               tokens:     _filters
            ],
            [  src:        "$_installConfigDir/Deployment/Metadata",
               include:    ~/.*\.orig/,
               dst:        _configMetadataDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        "$_installSasmiscDir/scenario",
               include:    ~/.*\.xml/,
               dst:        "$_configDataDir/scenario/xml",
               tokens:     _filters
            ],
         ]

         t__loadmetadata_createfolders =
         [
            _productFolder,
            _applicationFolder,
            _applicationSubFolder
         ]

         if (pass.hasVariable('_updateMetadata') && _updateMetadata)
            updateconfigure__loadmetadata_init()
         else t__loadmetadata_load = _appxmlFile

         t__importpackages_createfolders =
         [
            _applicationCodeFolder
         ]

         t__importpackages_run =
         [
            packages:
            [
               [
               name:          'ComplianceSolutions',
               packagesDir:   canonical("$_installConfigDir/Deployment/Packages/"),
               folder:        _applicationCodeFolder,
               tokens:        _filters
               ]
            ]
         ]
         // add message: with message from antimnycmnc.txt to properties
         t__addlogintogroup =
         [
            authdomains:
            [
               [
                  authdomain: _coreAuth,
                  login:      _coreUserId,
                  passwd:     api.'antimnycmnc.dbms.core.passwd',
                  group:      _coreGroup,
               ],
               [
                  authdomain: _kcAuth,
                  login:      _kcUserId,
                  passwd:     api.'antimnycmnc.dbms.kc.passwd',
                  group:      _kcGroup,
               ],
               [
                  authdomain: _segkcAuth,
                  login:      _segkcUserId,
                  passwd:     api.'antimnycmnc.dbms.segkc.passwd',
                  group:      _segkcGroup,
               ]
            ]
         ]

         t__sasjobs_run =
         [
            enabled: _initDB,
            folder:  _configInstallDir,
            groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
            jobs:
            [
               [
                  jobname: sasFCFInsJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobFCFInsFailed,
               ],
               [
                  jobname: sasFCFECMJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobFCFECMFailed,
               ],
               [
                  jobname: sasAMLInsJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobAMLInsFailed,
               ],
               [
                  jobname: sasAMLECMJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobAMLECMFailed,
               ],
               [
                  jobname: sasCDDInsJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobCDDInsFailed,
               ],
               [
                  jobname: sasCDDECMJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobCDDECMFailed,
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>
      <p>This call back is used when in migration and sets up a properties that will be used to move directories on disk.</p>

      @param t__movedirsondisk <type>Map&lt;ArrayList&gt;</type>: Datastructure that allows movement of directories on disk.<br>
            <indent><key>move</key>- <type>Arraylist</type>: Action to be performed on directory<br>
               <indent><key>src</key>- <type>String</type>: fully qualified path to source directory<br>
               <key>dst</key>- <type>String</type>: fully qualified path to destination directory, this directory must not exist<br>
               <key>newName</key>- <type>String</type> <kind>optional</kind>: name that the source directory is to be renamed to<br></indent></indent>
   */
   def configure__migrate_properties_post =
   {

      String sasFCFUpdteJob         = "fcf_update"
      String sasJobFCFUpdteFailed   = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasFCFUpdteJob],prefix:'${',suffix:'}').output[0]

      pass.with {
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]

         switch (_migrationFrom)
         {
         case '_51_and_earlier':
            break

         case '_52_and_61':
            if (_maintenanceFrom == '6.1')
            {
               t__movedirsondisk =
               [
                  move:
                  [
                     [ src: canonical("$_sasConfigDir/SASAntiMoneyLaundering6.1"),
                       dst: _configDir,
                       newName: 'SASAntiMoneyLaundering6.1'
                     ],
                     [ src: canonical("$_sasConfigDir/SASCustomerDueDiligence6.1"),
                       dst: _configDir,
                       newName: 'SASCustomerDueDiligence6.1'
                     ],
                     [ src: canonical("$_sasConfigDir/SASFinancialCrimesFoundation6.1"),
                       dst: _configDir,
                       newName: 'SASFinancialCrimesFoundation6.1'
                     ]
                  ]
               ]
               if (!_initDB)
               {
                  t__sasjobs_run =
                  [
                     enabled: true,
                     folder:  _configInstallDir,
                     groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
                     jobs:
                     [
                        [
                           jobname: sasFCFUpdteJob,
                           win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                           unx:     [sasexec:   _installUnxU8sasexec],
                           message: sasJobFCFUpdteFailed,
                        ]
                     ]
                  ]
               }
            }
            break

         case '_62_and_later':
            if (_maintenanceFrom == '6.2')
            {
               t__movedirsondisk =
               [
                  move:
                  [
                     [ src: canonical("$_sasConfigDir/SASComplianceSolutions6.2"),
                       dst: _configDir,
                       newName: '6.2'
                     ]
                  ]
               ]
               if (!_initDB)
               {
                  t__sasjobs_run =
                  [
                     enabled: true,
                     folder:  _configInstallDir,
                     groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
                     jobs:
                     [
                        [
                           jobname: sasFCFUpdteJob,
                           win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                           unx:     [sasexec:   _installUnxU8sasexec],
                           message: sasJobFCFUpdteFailed,
                        ]
                     ]
                  ]
               }
            }
            break

         default:
            break
         }
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>
   */
   def configure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.antimnycmnc.head',
                  value:api.'instructions_antimnycmnc_head.txt']
               ],
               [replace:[
                  token:'instructions.antimnycmnc.details.table.head',
                  value:api.'instructions_antimnycmnc_details_table_head.txt']
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_documentation.txt']
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_details_logging.txt']
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_details_configdir.txt']
               ],
               [replace:[
                  token:'instructions.antimnycmnc.details.table.tail',
                  value:api.'instructions_antimnycmnc_details_table_tail.txt']
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; migration</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>
   */
   def configure__docparts_migration =
   {
      pass.with {
         switch (_migrationFrom)
         {
            case '_52_and_61' :
               if (_maintenanceFrom == '5.2')
                  text = api.'instructions_antimnycmnc_migration_52.txt'
               else
                   text = api.'instructions_antimnycmnc_migration_61.txt'
               break
            case '_62_and_later' :
                  text = api.'instructions_antimnycmnc_migration.txt'
               break
            default:
               text = api.'instructions_antimnycmnc_migration_51_and_earlier.txt'
               break
         }
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.antimnycmnc.head',
                  value:api.'instructions_antimnycmnc_head.txt']
               ],
               [replace:[
                  token:'instructions.antimnycmnc.details.table.head',
                  value:api.'migration_antimnycmnc_details_table_head.txt']
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:text]
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_documentation.txt']
               ],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_details_configdir.txt']
               ],
               [replace:[
                  token:'instructions.antimnycmnc.details.table.tail',
                  value:api.'instructions_antimnycmnc_details_table_tail.txt']
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>sasjobs &gt; createuser</hook>
      <p>Calls method createUserAddToGroups and passes in the datastructure t__sasjobs_run</p>
   */
   def sasjobs_createuser =
   {
      api.createUserAddToGroups(pass.t__sasjobs_run)
   }

   /**
      Callback for hook: <hook>sasjobs &gt; createuser</hook>
      <p>Calls method removeUserFromGroupsAndDelete and passes in the datastructure t__sasjobs_run</p>
   */
   def sasjobs_removeuser =
   {
      api.removeUserFromGroupsAndDelete(pass.t__sasjobs_run)
   }

   // ===================================================== TARGET: updateConfigure

   /**
      Target: updateConfigure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def updateConfigure()
   {
      //api.stop()

      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      String maintenanceFrom = getProperty('antimnycmnc.maintenance.from.version', '')
      api.createProperty('antimnycmnc.maint.from.release', "${api.baseVersion(maintenanceFrom)}")

      if (maintenanceFrom)
      {
         String fromVersion         = api.baseVersion(maintenanceFrom)
         String maintenanceTo = getProperty('antimnycmnc.maintenance.to.version', '')
         String toVersion           = api.baseVersion(maintenanceTo)

         if (maintenanceFrom != maintenanceTo)
         {
            if (fromVersion != toVersion)
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > pre']            .add(updateconfigure__pre_init)
               t['properties > init']           .add(properties_init)
               t['properties > post']           .add(properties_post)
               t['properties > post']           .add(configure__properties_post)
               t['properties > post']           .add(updateconfigure__properties_post)
               t['moreproperties > init']       .add(updateconfigure__moreproperties_init)
               t['loadmetadata > init']         .add(updateconfigure__loadmetadata_init)
               t['searchandcall > init']        .add(updateconfigure__searchandcall_init)
               t['sasjobs > createuser']        .add(sasjobs_createuser)
               t['sasjobs > removeuser']        .add(sasjobs_removeuser)
               t['move > init']                 .add(updateconfigure__movedirsondisk_init)
               t['docparts > init']             .add(updateconfigure__docparts_init)

               t.execute()
            }
            else
            {
            //TODO
            // Check if this section is still required
            // it handles the maintenance for a specific release
            // 6.3 to 6.3_M1 etc.
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > init']           .add(properties_init)
               t['properties > post']           .add(properties_post)
               t['properties > post']           .add(updateconfigure__properties_post)
               t['moreproperties > init']       .add(updateconfigure__moreproperties_init)
               t['sasjobs > createuser']        .add(sasjobs_createuser)
               t['sasjobs > removeuser']        .add(sasjobs_removeuser)
               t['docparts > init']             .add(updateconfigure__docparts_init)

               t.execute()

            }
         }
      }
      else
         configure()

     //api.stop()
   }

   def updateconfigure__pre_init =
   {
      String legalName = api.createProperty('antimnycmnc.short.legal.name', 'Compliance Solutions Svr')

      String toVersion           = api.baseVersion(api.'antimnycmnc.maintenance.to.version')
      String fromVersion         = api.baseVersion(api.'antimnycmnc.maintenance.from.version')
      String releaseNumber       = api.'antimnycmnc.release.number'
      api.createProperty('antimnycmnc.release.number', toVersion)

      String softwareComponent   = api.createProperty('antimnycmnc.software.component.name', "$legalName $toVersion")
      String applicationFolder   = api.'antimnycmnc.system.product.folder'
      String oldFolderName       = "$applicationFolder/$legalName $fromVersion"

      String configDir = "${new File(canonical("${api.'config.lev.applications.dir'}/${api.'antimnycmnc.normalized.offering.legal.name'}"))}"
      api.createProperty('antimnycmnc.config.dir', configDir)

      String autoexecPath = canonical("${api.'config.lev.applications.dir'}/${api.'antimnycmnc.normalized.offering.legal.name'}/custom/config/fcf_autoexec.sas")
      api.createProperty('antimnycmnc.autoexec.path', autoexecPath)

      String subFolder           = api.createProperty('antimnycmnc.system.product.version.folder', "$applicationFolder/$softwareComponent")

      String midComponentName    = api.'antimnycmnc.offering.short.legal.name' + ' ' + 'Mid' + ' ' + toVersion

      api.createProperty('antimnycmnc.application.stp.folder', "$subFolder/Application SAS Code")
      api.createProperty('antimnycmnc.software.component.name', softwareComponent)
      api.createProperty('antimnycmnc.configobj.name', softwareComponent)
      api.createProperty('antimnycmnc.desc', softwareComponent)
      api.createProperty('antimnycmnc.antimnylndin.software.component.name', midComponentName)

      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : oldFolderName,
                  deleteContents : true
               ],
            ]
         ]
      }
   }

   def updateconfigure__properties_post =
   {
      String maintenanceFrom        = getProperty('antimnycmnc.maintenance.from.version', '')
      String maintenanceTo          = getProperty('antimnycmnc.maintenance.to.version', '')

      pass.with
      {

         _maintFromRelease       = api.baseVersion(maintenanceFrom)
         _maintFromLevel         = api.maintLevel(maintenanceFrom)
         _maintToRelease         = api.baseVersion(maintenanceTo)
         _maintToLevel           = api.maintLevel(maintenanceTo)

         _filters +=
         [
            'maint.from.release'       : _maintFromRelease,
            'maint.from.level'         : _maintFromLevel,
            'maint.to.release'         : _maintToRelease,
            'maint.to.level'           : _maintToLevel
         ]

         t__setobjectproperties =
         [
            properties:
            [
               'antimnycmnc.autoexec.path',
               'antimnycmnc.config.dir',
               'antimnycmnc.configobj.name'
            ]
         ]

         t__setconfigureversion =
         [
            fqid:       api.'antimnycmnc.configobj.fqid',
            version:    maintenanceTo
         ]
      }
   }

   def updateconfigure__moreproperties_init =
   {
      // SAS Jobs
      String sasFCFUpdteJob         = "fcf_update"
      String sasJobFCFUpdteFailed   = new PipeCode([api.'warnings_antimnycmnc_runsasjob_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasFCFUpdteJob],prefix:'${',suffix:'}').output[0]

      pass.with
      {
         t__sasjobs_run =
         [
            enabled: true,
            folder:  _configInstallDir,
            groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
            jobs:
            [
               [
                  jobname: sasFCFUpdteJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobFCFUpdteFailed,
               ]
            ]
         ]

         t__copyfiles +=
         [
            [  src:        _installDeployDir,
               include:    ~/fcf_update\.sas/,
               dst:        _configInstallDir,
               tokens:     _filters
            ]
         ]
      }
   }

   def updateconfigure__loadmetadata_init =
   {
      pass.with
      {
         t__loadmetadata_update =
         [
            file: canonical("$_configMetadataDir/antimnycmnc.appxml"),
            folder: _applicationSubFolder,
            applicationname: _softwareComponentName
         ]
      }
   }

   def updateconfigure__searchandcall_init =
   {
      String buName        = api.'antimnycmnc.business.unit.name'
      String dbmsType      = api.'antimnycmnc.dbms.type'
      String sasServerName = "FCF - $buName $dbmsType"

      pass.with
      {
         if (_maintFromRelease == 6.1)
         {
            t__searchandcall =
            [
               actions: [ __template.t__deletemetadata ],
               fqids:
               [
                  [
                     type:    'SASLibrary',
                     search:  "@Name contains 'FCF -'"
                  ],
                  [
                     type:    'SASClientConnection',
                     search:  "@Name contains 'Core' or @Name contains 'Knowledge Center'"
                  ],
                  [
                     type:    'ServerComponent',
                     search:  "@Name='$sasServerName'"
                  ],
               ]
            ]

            t__removemetadatabypath =
            [
               paths:
               [
                  [
                     objectPath:       '/System/Applications/SAS Anti-Money Laundering',
                     deleteContents:   true
                  ],
                  [
                     objectPath:       '/System/Applications/SAS Customer Due Diligence',
                     deleteContents:   true
                  ],
               ]
            ]
         }
      }
   }

   def updateconfigure__movedirsondisk_init =
   {
      pass.with
      {
         switch (_maintFromRelease)
         {
            case '6.2' :
               t__movedirsondisk =
               [
                  move:
                  [
                     [  src:        canonical("$_sasConfigDir/$_normalizedOfferingLegalName$_maintFromRelease"),
                        dst:        _configDir,
                        newName:    _maintFromRelease
                     ]
                  ]
               ]
               break
            case '6.1' :
               t__movedirsondisk =
               [
                  move:
                  [
                     [  src:        canonical("$_sasConfigDir/SASAntiMoneyLaundering$_maintFromRelease"),
                        dst:        _configDir
                     ],
                     [  src:        canonical("$_sasConfigDir/SASCustomerDueDilligence$_maintFromRelease"),
                        dst:        _configDir
                     ],
                     [  src:        canonical("$_sasConfigDir/SASFinancialCrimesFoundation$_maintFromRelease"),
                        dst:        _configDir
                     ]
                  ]
               ]
               break
            default:
               break
         }
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def updateconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  token:'instructions.antimnycmnc.head',
                  value:api.'instructions_antimnycmnc_head.txt']],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.head',
                  value:api.'instructions_antimnycmnc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'updateconfig_antimnycmnc_details_table_item_header.txt']],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'instructions_antimnycmnc_documentation.txt']],
               [prepend:[
                  token:'instructions.antimnycmnc.details.table.item',
                  value:api.'updateconfig_antimnycmnc_details_table_item_note.txt']],
               [replace:[
                  token:'instructions.antimnycmnc.details.table.tail',
                  value:api.'instructions_antimnycmnc_details_table_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: checkForErrors

   /**
      Target: checkForErrors

      @see com.sas.gate.api.template.CheckForErrorsTemplate
   */
   def checkForErrors()
   {
      CheckForErrorsTemplate t = new CheckForErrorsTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: reportFailures

   /**
      Target: reportFailures

      @see com.sas.gate.api.template.ReportFailuresTemplate
   */
   def reportFailures()
   {
      ReportFailuresTemplate t = new ReportFailuresTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: unconfigure

   /**
      Target: unconfigure

      @see com.sas.gate.api.template.UnconfigureTemplate
   */
   def unconfigure()
   {
      //api.stop()

      UnconfigureTemplate t = new UnconfigureTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['init']               .add(unconfigure__init)
      t['removemetadatabypath > init']  .add(unconfigure__removemetadatabypath_init)
      t['docparts > init']    .add(unconfigure__docparts_init)

      t.execute()

      //api.stop()
   }

   def unconfigure__removemetadatabypath_init =
   {
      pass.with
      {
         String path = (api.getmetadatavalues(
            type:       'SoftwareComponent',
            search:     "@Name='Compliance Svr LASR Cfg $_releaseNumber']" )) ? _applicationSubFolder : _applicationFolder

         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : path,
                  deleteContents : true
               ],
            ]
         ]
      }
   }

   def unconfigure__init =
   {
      pass.with
      {
         t__deleteauthdomain =
         [
            message: api.'antimnycmnc.status.dbms.auth.delete.failure.msg',
            authdomains:
            [
               api.'antimnycmnc.dbms.core.auth',
               api.'antimnycmnc.dbms.kc.auth',
               api.'antimnycmnc.dbms.segkc.auth'
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
            <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def unconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnycmnc.head',
                  value:   api.'unconfiguration_summary_antimnycmnc_head.txt']],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnycmnc.item',
                  value:   api.'unconfiguration_summary_antimnycmnc_item.txt']],
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.antimnycmnc.tail',
                  value:   api.'unconfiguration_summary_antimnycmnc_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: updatePasswords

   def updatePasswords()
   {
      //api.stop(properties:true)
      UpdatePasswordsTemplate t = new UpdatePasswordsTemplate(this)

      t['properties > init']        .add(properties_init)
      t['properties > post']        .add(properties_post)
      t['updatepasswords > init']   .add(updatepasswords__init)
      t['docparts > init']          .add(updatepasswords__docparts_init)

      t.execute()
      //api.stop(properties:true)
   }

   def updatepasswords__init =
   {

      pass.with {
         t__updatepasswords =
         [
            authdomains: [ // Note the authdomains must match the order in the antimnycmnc.credentials.persisted.list
                           api.'antimnycmnc.dbms.core.auth',
                           api.'antimnycmnc.dbms.kc.auth',
                           api.'antimnycmnc.dbms.segkc.auth'
                         ]
         ]
      }
   }

	/**
	 Callback for hook: <hook>updatepasswords &gt; docparts &gt; init</hook>
	 <p>This callback is used by the updatepasswords recipe to setup initial properties for the docparts.</p>

	 @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
	  <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
	  <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
	  <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
	  <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

	  Note: using the antimnylndcf tokes for update password this is not an error do not change!
	 */
   def updatepasswords__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            file:    api.'config.lev.updatepasswords.file',
            actions:
            [
               [replace:[
                  token:'updatePasswords.summary.antimnylndcf.head',
                  value:api.'updatePasswords_summary_antimnycmnc_head.txt']],
               [prepend:[
                  token:'updatePasswords.summary.antimnylndcf.item',
                  value:api.'updatePasswords_summary_antimnycmnc_kc_item.txt']],
               [prepend:[
                  token:'updatePasswords.summary.antimnylndcf.item',
                  value:api.'updatePasswords_summary_antimnycmnc_core_item.txt']],
               [prepend:[
                  token:'updatePasswords.summary.antimnylndcf.item',
                  value:api.'updatePasswords_summary_antimnycmnc_seg_kc_item.txt']],
               [replace:[
                  token:'updatePasswords.summary.antimnylndcf.tail',
                  value:api.'updatePasswords_summary_antimnycmnc_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: changeHost

   def changeHost()
   {
      //api.stop(properties:true)
      ChangeHostTemplate t = new ChangeHostTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['changehost > init']  .add(changehost__init)
      t['docparts > init']    .add(changehost__docparts_init)

      t.execute()
      //api.stop(properties:true)
   }

   def changehost__init =
   {
      pass.with {
         t__changehost =
         [
            exact: ['antimnycmnc.dbms.host'],
            updatefiles:
            [
               [
                  preToken: 'metaserver      = "',
                  file:     _autoexecPath,
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def changehost__docparts_init =
   {
      pass.with
      {
         if (t__changehost.docparts)
         {
            t__docparts =
            [
               file:    api.'config.lev.changehost.file',
               actions: []
            ]

            if (t__changehost.docparts.substring)
            {
               def (oldProperties, updatedProperties) = t__changehost.docparts.substring
               oldProperties.each { k, v ->

                  def value = new PipeCode([api.'changeHost_metadata_object_changed.txt'])
                     .tokenstringsub(tokens:
                     [
                        '_!PROPERTY_NAME':   k,
                        '_!OLD_VALUE':       v,
                        '_!NEW_VALUE':       updatedProperties."$k",
                     ], prefix:'${', suffix:'}').output[0]

                  t__docparts.actions += [prepend:[token: 'changeHost.configobj.item', value:value]]
               }
            }
         }
      }
   }
}
