// Copyright 2013-2014, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.

package com.sas.deployment
import groovy.transform.InheritConstructors

import com.sas.gate.CodeAccessor
import com.sas.gate.api.PipeCode
import com.sas.gate.api.template.*
import org.codehaus.groovy.runtime.StackTraceUtils

/**
   <overview>Deployment code for SAS Customer Due Diligence (hpamlsvrc)</overview>
   <p>
      This code configures the deployment of hpamlsvrc using what corresponds to
      the following Ant targets:
      <ul class='nodot'>
         <li>configure</li>
         <li>unconfigure</li>
      </ul>
   </p>

   @since      14w14
   @version    15w04
   @author     Bruce Bryan
   @see        #SOURCE#
*/

@InheritConstructors
class HPAMLSVRC extends CodeAccessor
{
   // ===================================================== GLOBAL

   /**
      Identifier

      @param id <type>String</type>: The 12byte name
   */
   String id = 'HPAMLSVRC'

   /**
      Callback for hook: <hook>properties &gt; init</hook>

      @param t__properties_get   <type>ArrayList&lt;String&gt;</type>: get object properties on each dependson property

   <ul class='nodot'>
      <li>levconfig</li>
      <li>casemgmtmvac</li>
      <li>antimnycmnc</li>
   </ul>
   */
   def properties_init =
   {
      pass.with {
         debug = ['debug']

         _sasConfigDir                       = api.'config.lev.applications.dir'
         _normalizedOfferingLegalName        = api.'hpamlsvrc.normalized.offering.name'
         _releaseNumber                      = api.'hpamlsvrc.release.number'
         _configDir                          = canonical("$_sasConfigDir/$_normalizedOfferingLegalName")
         _configInstallDir                   = canonical("$_configDir/install")
         _configMetadataDir                  = canonical("$_configInstallDir/metadata")
      // Metadata Folders (Folders and Items)
         _productsFolder                     = api.'hpamlsvrc.products.folder'
         _applicationFolder                  = api.'hpamlsvrc.application.folder'
         _applicationSubFolder               = api.'hpamlsvrc.application.sub.folder'
         _applicationCodeFolder              = "$_applicationSubFolder/Application SAS Code"
         _softwareComponentName              = api.'hpamlsvrc.software.component.name'

         t__properties_get =
         [
            'hpamlsvrc.dependson.levconfig.fqid',
            'hpamlsvrc.dependson.casemgmtmvac.fqid',
            'hpamlsvrc.dependson.antimnycmnc.fqid',
            'hpamlsvrc.dependson.bihpgrdc.fqid',
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param _filters <type>Map&lt;String&gt;</type>: tokens for string substitution
      @param t__setobjectproperties    <type>ArrayList&lt;String&gt;</type>: Properties that will be stored in the configured objects text store
      @param t__copyfiles_createdirectories <type>Map&lt;String&gt;</type>: directories to be created in configure target<br>
            <indent><key>prefix</key>- <type>String</type>: base directory to create<br>
            <key>entries</key>- <type>ArrayList&lt;String&gt;</type>: directories and subdirectories to be created<br></indent>
      <pre>
   /install/config/Lev1/Applications/SASHighPerformanceAntiMoneyLaundering
`-- 6.3
    |-- custom
    |   |-- config
    |   |-- fmts
    |   `-- source
    |-- data
    |   |-- master
    |   |   `-- alert
    |   `-- stage
    |       `-- alert
    |-- install
    |   |-- metadata
    |   `-- xml
    |-- lasr
    |-- sasmisc
    `-- scenario
        `-- scenario_source</pre>

      @param t__copyfiles         <type>ArrayList&lt;Map&gt;</type>: files to copy and conditions for the copy<br>
            <indent><key>src</key>- <type>String</type>: Source directory to copy from<br>
            <key>include</key>- <type>String</type>: Regex pattern of file(s) to copy<br>
            <key>dst</key>- <type>String</type>: Destination directory for files to be copied to<br>
            <key>strip</key>- <type>String</type> <kind>optional</kind>: Remove extension form file name i.e., remove .orig extension<br>
            <key>tokens</key>- <type>String</type> <kind>optional</kind>: filterset to use for token replacement<br></indent>
      @param t__loadmetadata_createfolders   -
      @param t__importpackages_createfolders   -
      @param t__loadmetadata_load   <type>String</type>: Fully qualified path to the token substituted appxml file
      @param t__sasjobs_run      -
   */
   def properties_post =
   {
      // Simple Strings
      String offeringLegalName                  = api.'hpamlsvrc.offering.legal.name'
      String legalName                          = api.'hpamlsvr.legal.name'
      String shortLegalName                     = api.'hpamlsvrc.short.legal.name'
      String offeringShortName                  = api.'hpamlsvrc.offering.short.name'

      // Config Paths
      String amlConfigDir                       = canonical("${api.'config.lev.applications.dir'}/SASComplianceSolutions")
      String amlAutoexecPath                    = canonical("${amlConfigDir}/custom/config/aml_autoexec.sas")
      String ecmConfigDir                       = canonical("${api.'config.lev.applications.dir'}/SASEnterpriseCaseManagement")
      String ecmAutoexecPath                    = canonical("${ecmConfigDir}/Source/control/ecm_autoexec.sas")
      String fcfAutoexecPath                    = canonical("${amlConfigDir}/custom/config/fcf_autoexec.sas")
      String fcfConfigDir                       = amlConfigDir
      String configCustFmtDir                   = canonical("$pass._configDir/custom/fmts")

      // Install Paths (Dirs and Files)
      String installSASDir                      = api.'sasinstall.install.dir'
      String installSASConfigFile               = canonical("$installSASDir/sasv9.cfg")
      String installSASExecFile                 = canonical("$installSASDir/sas.exe")
      String installSasmiscDir                  = api.'hpamlsvrc.install.sasmisc.dir'
      String installCmacrosDir                  = api.'hpamlsvrc.install.cmacros.dir'
      String installUcmacrosDir                 = api.'hpamlsvrc.install.ucmacros.dir'

      // Other
      String amlDavPath                         = api.'antimnycmnc.dav.path'
      String classIdentifier                    = api.'hpamlsvrc.configobj.classid'
      String currProd12byte                     = api.'config.currprod.12byte'
      String gridHost                           = api.'bihpgrdc.lasrserver.host'
      String gridPath                           = api.'bihpgrdc.lasrserver.path'
      String gridPort                           = 0
      String majorNumber                        = api.'hpamlsvrc.major.number'
      String minorNumber                        = api.'hpamlsvrc.minor.number'
      String nefs                               = File.separatorChar
      String segkcUserID                        = api.'antimnycmnc.dbms.segkc.userid'
      String serverContextName                  = api.'oma.appserver.servercontext.name'

      // Properties
      api.createProperty('hpamlsvrc.config.dir', pass._configDir)
      api.createProperty('hpamlsvrc.configobj.name', pass._softwareComponentName)

      if (t__debug){
         assert offeringLegalName == 'SAS High-Performance Anti-Money Laundering' : "Offering legal name does not match: SAS High-Performance Anti-Money Laundering"
         assert legalName == 'SAS High-Performance Anti-Money Laundering Server' : "Offering legal name does not match: SAS High-Performance Anti-Money Laundering Server"
         assert shortLegalName == 'High-Perf AntiMoneyLaund Sv' : "Short legal name does not match: High-Perf AntiMoneyLaund Sv"
         assert offeringShortName == 'High-Perf AML' : "Offering short legal name does not match: High-Perf AML"
         assert majorNumber == '6' : "Release major number does not match: 6"
         assert minorNumber == '3' : "Release minor number does not match: 3"
      }

      pass.with
      {
         _initDB                             = !(getProperty('antimnycmnc.bypass_init', '0') { toBoolean(it) })
         _installSampleDir                   = api.'hpamlsvrc.install.sample.dir'
         _installStpDir                      = api.'hpamlsvrc.install.sasstp.dir'
         _installConfigDir                   = api.'hpamlsvrc.install.config.dir'
         _installSasmiscDir                  = api.'hpamlsvrc.install.sasmisc.dir'
         _installDeployDir                   = canonical("$_installSasmiscDir/deploy")
         _installMetadataDir                 = canonical("$_installConfigDir/Deployment/Metadata")
         _installSASNLSConfigFile            = canonical("$installSASDir/nls/u8/sasv9.cfg")
         _installUnxU8sasexec                = canonical("$installSASDir/bin/sas_u8")
         _configCustSrcDir                   = canonical("$_configDir/custom/source")
         _configDataDir                      = canonical("$_configDir/data")
         _configStpDir                       = canonical("$_configDir/sasstp")
         _configCustCfgDir                   = canonical("$_configDir/custom/config")
         _autoexecPath                       = canonical("$_configCustCfgDir/hpaml_autoexec.sas")
         _pgaAutoexecPath                    = canonical("$_configCustCfgDir/pga_autoexec.sas")
         _coreGroup                          = 'Core'
         _kcGroup                            = 'Knowledge Center'
         _segkcGroup                         = 'Business Unit Knowledge Center'

         _filters =
         [
            'aml.autoexec.path'                    : amlAutoexecPath,
            'aml.dav.path'                         : amlDavPath,
            'appserver.servercontext.name'         : serverContextName,
            'class.identifier'                     : classIdentifier,
            'custsrc'                              : _configCustSrcDir,
            'description'                          : _softwareComponentName,
            'ecm.autoexec.path'                    : ecmAutoexecPath,
            'fcf.autoexec.path'                    : fcfAutoexecPath,
            'grid.host'                            : gridHost,
            'grid.path'                            : gridPath,
            'grid.port'                            : gridPort,
            'hpaml.autoexec.path'                  : _autoexecPath,
            'hpaml.custfmt'                        : configCustFmtDir,
            'hpaml.install.misc.dir'               : _installSasmiscDir,
            'hpaml.lasr.host'                      : gridHost,
            'hpaml.lasr.path'                      : gridPath,
            'hpamldata'                            : _configDataDir,
            'hpamlroot'                            : _configDir,
            'legal.name'                           : legalName,
            'major.number'                         : majorNumber,
            'minor.number'                         : minorNumber,
            'nefs'                                 : nefs,
            'offering.legal.name'                  : offeringLegalName,
            'short.legal.name'                     : shortLegalName,
            'pga.autoexec.path'                    : _pgaAutoexecPath,
            'pga.custsrc'                          : _configCustSrcDir,
            'pga.dav.path'                         : amlDavPath,
            'pga.install.cmacros.dir'              : installCmacrosDir,
            'pga.install.ucmacros.dir'             : installUcmacrosDir,
            'pga.software.component.name'          : _softwareComponentName,
            'pgadata'                              : _configDataDir,
            'pgaroot'                              : _configDir,
            'product.version.folder'               : _applicationSubFolder,
            'release.number'                       : _releaseNumber,
            'stproot'                           	: _configStpDir.replaceAll('\\\\', '\\\\\\\\'),
            'segment_id'                           : segkcUserID,
            'software.component.name'              : _softwareComponentName
         ]
      }
      api.createProperty('hpamlsvrc.autoexec.path', pass._autoexecPath)

   }

   def properties_assert =
   {
      pass.with {
         assert _normalizedOfferingLegalName == 'SASHighPerformanceAntiMoneyLaundering' : "Normalized offering legal name does not match: SASHighPerformanceAntiMoneyLaundering"
         assert _releaseNumber == '6.3' : "SAS High-Performance Anti-Money Laundering release number does not match: 6.3"
         assert _configDir =~ "SASHighPerformanceAntiMoneyLaundering" : "Config dir does not contain: SASHighPerformanceAntiMoneyLaundering"
         assert _applicationFolder == '/System/Applications/SAS High-Performance Anti-Money Laundering' : "Application metadata folder does not match: /System/Applications/SAS High-Performance Anti-Money Laundering"
         assert _applicationSubFolder =~ 'High-Perf AntiMoneyLaund Sv 6.3' : "Application metadata sub folder does not contain: High-Perf AntiMoneyLaund Sv 6.3"
         assert _applicationCodeFolder =~ 'Application SAS Code' : "Application metadata SAS code folder does not contain: Application SAS Code"
         assert _softwareComponentName == 'High-Perf AntiMoneyLaund Sv 6.3' : "Software component name does not match: High-Perf AntiMoneyLaund Sv 6.3"
      }
   }

   // ===================================================== TARGET: configure

   /**
      Target: configure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def configure()
   {
      //api.stop()

      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      ConfigureTemplate t = new ConfigureTemplate(this)

      t['migration > init']      .add(configure__migration_init)
      t['properties > init']     .add(properties_init)
      t['properties > post']     .add(properties_post)
      t['properties > post']     .add(configure__properties_post)
      t['properties > post']     .add(configure__migrate_properties_post).when('t__migrate')
      t['properties > post']     .add(properties_assert).when('t__debug')
      t['docparts > init']       .add(configure__docparts_init).when('!t__migrate')
      t['docparts > migration']  .add(configure__docparts_migration).when('t__migrate')

      t['sasjobs > createuser']  .add(sasjobs_createuser).when('!t__migrate')
      t['sasjobs > removeuser']  .add(sasjobs_removeuser).when('!t__migrate')

      t.execute()

      //api.stop()
   }

   def configure__migration_init =
   {
      pass.with {
         _updateMetadata =
         t__migrate = getProperty('hpamlsvrc.migration.is_enabled', '0') { toBoolean(it) }
      }
   }

   def configure__properties_post =
   {
      // SAS Job
      String sasHPECMJob                        = 'hpaml_ecm_config'
      String sasHPInsJob                        = 'hpaml_install'
      String sasPGAECMJob                       = 'pga_ecm_config'
      String sasPGAInsJob                       = 'pga_install'
      String sasJobHPECMFailed                    = new PipeCode([api.'warnings_hpamlsvrc_loadInstallData_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasHPECMJob],prefix:'${',suffix:'}').output[0]
      String sasJobHPInsFailed                  = new PipeCode([api.'warnings_hpamlsvrc_loadInstallData_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasHPInsJob],prefix:'${',suffix:'}').output[0]
      String sasJobPGAECMFailed                 = new PipeCode([api.'warnings_hpamlsvrc_loadInstallData_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasPGAECMJob],prefix:'${',suffix:'}').output[0]
      String sasJobPGAInsFailed                 = new PipeCode([api.'warnings_hpamlsvrc_loadInstallData_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasPGAInsJob],prefix:'${',suffix:'}').output[0]

      pass.with
      {
         t__createdirectories =
         [
            prefix: _configDir,
            entries:
            [
               'custom/config',
               'custom/source',
               'custom/fmts',
               'data/master',
               'data/master/alert',
               'data/stage',
               'data/stage/alert',
               'install/metadata',
               'lasr',
               'sasmisc',
               'sasstp',
               'scenario/scenario_source',
            ]
         ]

         t__setobjectproperties =
         [
            properties:
            [
               'hpamlsvrc.release.number',
               'hpamlsvrc.software.component.name',
            ]
         ]

         t__copyfiles =
         [
            [  src:        _installSampleDir,
               include:    ~/.*\.(sas|cmd)/,
               exclude:    ~/(autoexec\.sas|install\.sas|hpaml_sa_load_data\.sas)/,
               dst:        _configCustSrcDir,
               tokens:     _filters
            ],
            [  src:        _installSampleDir,
               include:    ~/.*\.sas7bdat/,
               dst:        _configDataDir
            ],
            [  src:        _installStpDir,
               include:    ~/.*\.(sas|orig)/,
               dst:        _configStpDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installDeployDir,
               recurse:    true,
               include:    ~/.*/,
               exclude:    ~/.*autoexec\.sas/,
               dst:        _configInstallDir,
               tokens:     _filters
            ],
            [  src:        _installDeployDir,
               include:    ~/.*autoexec\.sas/,
               dst:        _configCustCfgDir,
               tokens:     _filters
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.orig/,
               dst:        _configMetadataDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.locprop/,
               dst:        _configMetadataDir
            ],
            [  src:        "$_installSasmiscDir/scenario/scenario_source",
               include:    ~/.*\.sas/,
               dst:        "$_configDir/scenario/scenario_source",
               tokens:     _filters
            ],
         ]

         t__loadmetadata_createfolders =
         [
            _applicationFolder,
            _applicationSubFolder,
            _productsFolder
         ]

         t__loadmetadata_load = canonical("$_configMetadataDir/hpamlsvrc.appxml")

         if (pass.hasVariable('_updateMetadata') && _updateMetadata)
            updateconfigure__loadmetadata_init()

         t__importpackages_createfolders =
         [
            _applicationCodeFolder
         ]

         t__importpackages_run =
         [
            packages:
            [
               [
               name:          'PeerGroupAnalysis',
               packagesDir:   canonical("$_installConfigDir/Deployment/Packages/"),
               folder:        _applicationCodeFolder,
               tokens:        _filters
               ]
            ]
         ]

         t__sasjobs_run =
         [
            enabled: _initDB,
            folder:  "$_configDir/install",
            groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
            jobs:
            [
               [
                  jobname: sasHPInsJob,
                  message: sasJobHPInsFailed,
               ],
               [
                  jobname: sasHPECMJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobHPECMFailed,
               ],
               [
                  jobname: sasPGAInsJob,
                  message: sasJobPGAInsFailed,
               ],
               [
                  jobname: sasPGAECMJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobPGAECMFailed,
               ]
            ]
         ]
      }
   }

   def configure__migrate_properties_post =
   {
      pass.with {
         _maintenanceFrom = api.'hpamlsvrc.maintenance.from.version'
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]

         if ( _releaseNumber != _maintenanceFrom )
         {
            t__removemetadatabypath =
            [
               paths:
               [
                  [
                     deleteContents : true,
                     objectPath : "/System/Applications/${api.'hpamlsvrc.offering.legal.name'}/${api.'hpamlsvrc.short.legal.name'} $_maintenanceFrom"
                  ]
               ]
            ]
         }
         else

         switch (_maintenanceFrom)
         {
         case '6.1':
            t__movedirsondisk =
            [
               move:
               [
                  [ src: canonical("$_sasConfigDir/SASPeerGroupAnalysis6.1"),
                    dst: _configDir,
                    newName: 'SASPeerGroupAnalysis6.1'
                  ]
               ]
            ]

            t__removemetadatabypath =
            [
               paths:
               [
                  [
                     deleteContents : true,
                     objectPath : "/System/Applications/SAS Peer Group Analysis"
                  ]
               ]
            ]
            break
         case '6.2':
            t__movedirsondisk =
            [
               move:
               [
                  [ src: canonical("$_sasConfigDir/SASHighPerformanceAntiMoneyLaundering6.2"),
                    dst: _configDir,
                    newName: '6.2'
                  ]
               ]
            ]
            break
         default:
            break
         }
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def configure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.hpamlsvrc.head',
                  value:api.'instructions_hpamlsvrc_head.txt']],
               [replace:[
                  token:'instructions.hpamlsvrc.details.table.head',
                  value:api.'instructions_hpamlsvrc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'instructions_hpamlsvrc_doc.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'instructions_hpamlsvrc_details_logging.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'instructions_hpamlsvrc_details_configdir.txt']],
               [replace:[
                  token:'instructions.hpamlsvrc.details.table.tail',
                  value:api.'instructions_hpamlsvrc_details_table_tail.txt']],
            ]
         ]
      }
   }

   def configure__docparts_migration =
   {
      pass.with
      {
         switch (_migrationFrom)
         {
            case '6.1' :
               text = api.'instructions_hpamlsvrc_migration_61.txt'
               break
            default:
               text = api.'instructions_hpamlsvrc_migration.txt'
               break
         }
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.hpamlsvrc.head',
                  value:api.'instructions_hpamlsvrc_head.txt']],
               [replace:[
                  token:'instructions.hpamlsvrc.details.table.head',
                  value:api.'migration_hpamlsvrc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:text]],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'instructions_hpamlsvrc_doc.txt']],
               [replace:[
                  token:'instructions.hpamlsvrc.details.table.tail',
                  value:api.'instructions_hpamlsvrc_details_table_tail.txt']],
            ]
         ]
      }
   }

   def sasjobs_createuser =
   {
      api.createUserAddToGroups(pass.t__sasjobs_run)
   }

   def sasjobs_removeuser =
   {
      api.removeUserFromGroupsAndDelete(pass.t__sasjobs_run)
   }

   // ===================================================== TARGET: updateConfigure

   /**
      Target: updateConfigure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def updateConfigure()
   {
      //api.stop()

      pass.with {
         t__debug = getProperty('wizard.debug.mode', '0') { toBoolean(it) }
      }

      def maintenanceFrom = getProperty("hpamlsvrc.maintenance.from.version", '')
      api.createProperty('hpamlsvrc.maint.from.release', api.baseVersion(maintenanceFrom))

      if (maintenanceFrom)
      {
         String fromVersion         = api.baseVersion(api.'hpamlsvrc.maintenance.from.version')
         String toVersion           = api.baseVersion(api.'hpamlsvrc.maintenance.to.version')
         String maintenanceTo = getProperty("hpamlsvrc.maintenance.to.version", '')

         if (maintenanceFrom != maintenanceTo)
         {
            if (fromVersion != toVersion)
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > pre']            .add(updateconfigure__pre_init)
               t['properties > init']           .add(properties_init)
               t['properties > post']           .add(properties_post)
               t['properties > post']           .add(configure__properties_post)
               t['properties > post']           .add(updateconfigure__properties_post)
               t['properties > post']           .add(properties_assert).when('t__debug')
               t['moreproperties > init']       .add(updateconfigure__moreproperties_init)
               t['loadmetadata > init']         .add(updateconfigure__loadmetadata_init)
               t['sasjobs > createuser']        .add(sasjobs_createuser)
               t['sasjobs > removeuser']        .add(sasjobs_removeuser)
               t['setobjectproperties > init']  .add(updateconfigure__setobjectproperties_init)
               t['docparts > init']             .add(updateconfigure__docparts_init)

               t.execute()
            }
            else
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > init']           .add(properties_init)
               t['properties > post']           .add(properties_post)
               t['properties > post']           .add(updateconfigure__properties_post)
               t['moreproperties > init']       .add(updateconfigure__moreproperties_init)
               t['sasjobs > createuser']        .add(sasjobs_createuser)
               t['sasjobs > removeuser']        .add(sasjobs_removeuser)
               t['docparts > init']             .add(updateconfigure__docparts_init)

               t.execute()
            }
         }
      }
      else
         configure()

     //api.stop()
   }

   def updateconfigure__pre_init =
   {
      String shortLegalName = api.createProperty('hpamlsvrc.short.legal.name', 'High-Perf AntiMoneyLaund Sv')

      String toVersion           = api.baseVersion(api.'hpamlsvrc.maintenance.to.version')
      String fromVersion         = api.baseVersion(api.'hpamlsvrc.maintenance.from.version')
      String releaseNumber       = api.'hpamlsvrc.release.number'
      api.createProperty('hpamlsvrc.release.number', toVersion)

      String softwareComponent   = api.createProperty('hpamlsvrc.software.component.name', "$shortLegalName $toVersion")
      String applicationFolder   = api.'hpamlsvrc.application.folder'
      String oldFolderName       = "$applicationFolder/$shortLegalName $fromVersion"

      String configDir= "${api.'config.lev.applications.dir'}/${api.'hpamlsvrc.normalized.offering.name'}"
      api.createProperty('hpamlsvrc.config.dir', configDir)
      String subFolder = api.createProperty('hpamlsvrc.application.sub.folder', "$applicationFolder/$softwareComponent")

      api.createProperty('hpamlsvrc.application.stp.folder', "$subFolder/Application SAS Code")

      pass.with
      {
         if (_maintFromLevel == 6.1)
         {
            t__removemetadatabypath =
            [
               paths:
               [
                  [
                     objectPath     : oldFolderName,
                     deleteContents : true
                  ],
                  [
                     objectPath:       '/System/Applications/SAS Peer Group Analysis',
                     deleteContents:   true
                  ],
               ]
            ]
         }
         else
         {
            t__removemetadatabypath =
            [
               paths:
               [
                  [
                     objectPath     : oldFolderName,
                     deleteContents : true
                  ]
               ]
            ]
         }
      }
   }

   def updateconfigure__properties_post =
   {
      // SAS Jobs
      String maintenanceFrom        = getProperty('hpamlsvrc.maintenance.from.version', '')
      String maintenanceTo          = getProperty('hpamlsvrc.maintenance.to.version', '')

      pass.with
      {
         _maintFromRelease       = api.baseVersion(maintenanceFrom)
         _maintFromLevel         = api.maintLevel(maintenanceFrom)
         _maintToRelease         = api.baseVersion(maintenanceTo)
         _maintToLevel           = api.maintLevel(maintenanceTo)

         _filters <<
         [
            'maint.from.release'       : _maintFromRelease,
            'maint.from.level'         : _maintFromLevel,
            'maint.to.release'         : _maintToRelease,
            'maint.to.level'           : _maintToLevel
         ]

         t__setconfigureversion =
         [
            fqid:       api.'hpamlsvrc.configobj.fqid',
            version:    maintenanceTo
         ]
      }
   }

   def updateconfigure__moreproperties_init =
   {
      // SAS Jobs
      String sasHPUpdteJob         = "hpaml_update"
      String sasJobHPUpdteFailed   = new PipeCode([api.'warnings_hpamlsvrc_loadInstallData_error.txt'])
         .tokenstringsub(tokens:['_!JOB_NAME':sasHPUpdteJob],prefix:'${',suffix:'}').output[0]

      pass.with {
         t__sasjobs_run =
         [
            enabled: true,
            folder:  _configInstallDir,
            groups:  [ 'Ent Case Mgmt DB Group', _coreGroup, _kcGroup, _segkcGroup ],
            jobs:
            [
               [
                  jobname: sasHPUpdteJob,
                  win:     [options:   '-CONFIG "' + _installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   _installUnxU8sasexec],
                  message: sasJobHPUpdteFailed,
               ]
            ]
         ]

         t__copyfiles <<
         [
            [  src:        _installDeployDir,
               include:    ~/hpaml_update\.sas/,
               dst:        _configInstallDir,
               tokens:     _filters
            ]
         ]

      }
   }

   def updateconfigure__loadmetadata_init =
   {
      pass.with {
         t__loadmetadata_update =
         [
            file: canonical("$_configMetadataDir/hpamlsvrc.appxml"),
            folder: _applicationSubFolder,
            applicationname: _softwareComponentName
         ]
      }
   }

   def updateconfigure__setobjectproperties_init =
   {
      pass.with {
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def updateconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  token:'instructions.hpamlsvrc.head',
                  value:api.'instructions_hpamlsvrc_head.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.head',
                  value:api.'instructions_hpamlsvrc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'updateconfig_hpamlsvrc_details_table_item_header.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'updateconfig_hpamlsvrc_details_table_item_note.txt']],
               [prepend:[
                  token:'instructions.hpamlsvrc.details.table.item',
                  value:api.'instructions_hpamlsvrc_doc.txt']],
               [replace:[
                  token:'instructions.hpamlsvrc.details.table.tail',
                  value:api.'instructions_hpamlsvrc_details_table_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: checkForErrors

   /**
      Target: checkForErrors

      @see com.sas.gate.api.template.CheckForErrorsTemplate
   */
   def checkForErrors()
   {
      CheckForErrorsTemplate t = new CheckForErrorsTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: reportFailures

   /**
      Target: reportFailures

      @see com.sas.gate.api.template.ReportFailuresTemplate
   */
   def reportFailures()
   {
      ReportFailuresTemplate t = new ReportFailuresTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: unconfigure

   /**
      Target: unconfigure

      @see com.sas.gate.api.template.UnconfigureTemplate
   */
   def unconfigure()
   {
      //api.stop()

      UnconfigureTemplate t = new UnconfigureTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['init']               .add(unconfigure__init)
      t['docparts > init']    .add(unconfigure__docparts_init)

      t.execute()

      //api.stop()
   }

   def unconfigure__init =
   {
      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : _applicationFolder,
                  deleteContents : true
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
            <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def unconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.hpamlsvrc.head',
                  value:   api.'unconfiguration_summary_hpamlsvrc_head.txt']],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.hpamlsvrc.item',
                  value:   api.'unconfiguration_summary_hpamlsvrc_item.txt']],
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.hpamlsvrc.tail',
                  value:   api.'unconfiguration_summary_hpamlsvrc_tail.txt']],
            ]
         ]
      }
   }
}
